import h5py
from .css_logger import log
import numpy as np
from typing import TypeVar


import os
import re


DATA_LOCATION = "/entry/data/"

T = TypeVar("T", bound=object)  # TODO: Bind to h5py stuff


def get(object, path: str, expected_type: type[T]) -> T:
    value = object[path]
    if not isinstance(value, expected_type):
        raise ValueError(f"Expected {expected_type} at {path}, got {type(value)}")
    return value


class H5Reader(object):

    def __init__(self, filename):
        self.filename = filename

    def get_nb_images(self):
        with h5py.File(self.filename) as ifd:
            return ifd[DATA_LOCATION].attrs["nb_images"]

    def get_image(self, imgno):
        with h5py.File(self.filename) as ifd:
            nb_images = ifd[DATA_LOCATION].attrs["nb_images"]
            if imgno > 0 and imgno <= nb_images:
                dgroup = get(ifd, DATA_LOCATION, expected_type=h5py.Group)

                for ky in dgroup.keys():
                    d = get(dgroup, ky, expected_type=h5py.Dataset)
                    fimg = d.attrs["image_nr_low"]
                    limg = d.attrs["image_nr_high"]
                    if imgno >= fimg and imgno <= limg:
                        idx = imgno - fimg
                        return d[()][idx]

        return None


class H5Writer(object):
    def __init__(self, filename, images_per_file=5, overwrite=False):
        self.filename = filename
        self.overwrite = overwrite
        self.nr_imgs_per_file = images_per_file

    def create(self):
        if not self.overwrite:
            self.filename = choose_filename(self.filename)

        log.log(2, "Creating file: %s" % self.filename)

        with h5py.File(self.filename, "w") as ofd:
            ofd.create_group(DATA_LOCATION)

    def save_metadata(self, metadata, rootky="/entry"):
        self.metadata = metadata
        with h5py.File(self.filename, "r+") as fd:
            self._save_metadata(fd, metadata, rootky)

    def _save_metadata(self, fd, values, rootky="/entry"):
        for ky, value in values.items():
            h5ky = os.path.join(rootky, ky)

            if isinstance(value, dict):
                try:
                    fd.create_group(h5ky)
                except:
                    pass
                self._save_metadata(fd, value, h5ky)
            else:
                print("looking for %s in %s" % (ky, rootky))
                if ky in fd[rootky].keys():
                    print("%s already exists" % ky)
                    if isinstance(value, str):
                        del fd[h5ky]
                        fd.create_dataset(h5ky, data=value)
                    else:
                        fd[h5ky].write_direct(np.array([value]))
                else:
                    print("%s creating" % ky)
                    if isinstance(value, str):
                        dtype = "S"
                    else:  # valid for single number values and for list/tuple objects
                        value = np.array(value)
                        dtype = value.dtype
                    fd.create_dataset(h5ky, data=value)

    def save_img(self, data):
        dkeys = []

        fileno = 0
        with h5py.File(self.filename, "r+") as ofd:
            data_group = get(ofd, DATA_LOCATION, expected_type=h5py.Group)

            for ky in data_group.keys():
                fileno += 1
                dkeys.append(ky)
                last_imgno = data_group[ky].attrs["image_nr_high"]

            # how many images in last one
            newfile = False
            if dkeys:
                dataky = os.path.join(DATA_LOCATION, dkeys[-1])
                odata = get(ofd, dataky, expected_type=h5py.Dataset)[()]
                nb_images = len(odata)
                if nb_images >= self.nr_imgs_per_file:
                    newfile = True
                    fileno += 1
                imgno = last_imgno + 1
            else:
                newfile = True
                fileno = 1
                imgno = 1

            dataky = "data_%06d" % fileno
            fullky = os.path.join(DATA_LOCATION, dataky)

            if newfile:
                data_filename = self.create_datafile(dataky, data=data)
                ofd[fullky] = h5py.ExternalLink(
                    os.path.basename(data_filename), os.path.join(DATA_LOCATION, "data")
                )
                ofd[fullky].attrs.create("image_nr_low", imgno)
                ofd[fullky].attrs.create("image_nr_high", imgno)
            else:  # append data to latest file
                dset = get(ofd, fullky, expected_type=h5py.Dataset)
                newlen = dset.shape[0] + 1
                newshape = (newlen, dset.shape[1], dset.shape[2])
                dset.resize(newshape)
                dset[newlen - 1] = data
                ofd[fullky].attrs.modify("image_nr_high", imgno)
            ofd[DATA_LOCATION].attrs.modify("nb_images", imgno)

    def create_datafile(self, dataky, data):
        data_filename = self.filename.replace("master", dataky)
        log.log(2, " - creating datafile %s" % data_filename)
        with h5py.File(data_filename, "w") as dfd:
            maxshape = tuple((None,) + data.shape)
            shape = tuple((1,) + data.shape)
            dfd.create_dataset(
                os.path.join(DATA_LOCATION, "data"),
                data=data,
                shape=shape,
                dtype=data.dtype,
                maxshape=maxshape,
            )
        return data_filename


def choose_filename(filename):
    if not os.path.exists(filename):
        return filename

    # search for _master
    mat = re.search(r"(.*?)(([\-\.\_]master)(.*?)$)", filename)
    if mat:
        base = mat.group(1)
        master = mat.group(2)
    else:
        base, master = os.path.splitext(filename)
        master = "." + master

    fileno = 1
    while os.path.exists(filename):
        filename = "%s_%02d%s" % (base, fileno, master)
        fileno += 1

    return filename


def test():
    import time
    import numpy as np

    metadata = {
        "instrument": {
            "chi": 23,
            "phi": 12,
            "th": 100,
            "scanpositions": [7, 2, 3, 4],
        },
        "save_time": time.time(),
        "detector": {
            "exptime": 1.1,
            "threshold": 111,
            "acqmode": "PC",
            "name": "StreakCamera",
        },
    }

    data = np.zeros((100, 100), dtype="f")

    w = H5Writer("/tmp/multi_master.h5")
    w.create()
    w.save_metadata(metadata)
    for i in range(23):
        d = data + i
        w.save_img(d)


if __name__ == "__main__":
    log.start()
    test()





import logging
from typing import Any, Callable, Generic, Literal, TypeVar, Type, Union

from pyee.asyncio import AsyncIOEventEmitter

from pyspec._connection.data import DataType

LOGGER = logging.getLogger(__name__)


T = TypeVar("T", bound=DataType)


class PropertyEventEmitter(Generic[T], AsyncIOEventEmitter):
    """
    This class extends AsyncIOEventEmitter to provide type-safe event emission and handling for property change events.
    """

    def emit(self, event: str, *args: Any, **kwargs: Any) -> None:  # type: ignore
        # type: (Literal["change"], T) -> None # type: ignore
        super().emit(event, *args, **kwargs)

    def on(self, event: Literal["change"], func: Callable[[T], Any]) -> Any:  # type: ignore
        super().on(event, func)


class Property(PropertyEventEmitter[T]):
    """
    Defines a property that can be remotely accessed by clients.

    Args:
        name (str): The name of the property.
        initial_value (T): The initial value of the property.
        dtype (type[T] | type[object], optional): The expected data type of the property. Defaults to object (no validation).
    """

    def __init__(
        self,
        name: str,
        initial_value: T,
        dtype: Union[Type[T], Type[object]] = object,
    ):
        super().__init__()
        self.name = name
        self._value: T = initial_value
        self._dtype = dtype

    def get(self) -> T:
        """
        Get the current value of the property.

        Returns:
            T: The current value of the property.
        """
        return self._value

    def set(self, value: T) -> None:
        """
        Set the value of the property and emit a change event.

        :param value: The new value to set.
        """
        if not isinstance(value, self._dtype):
            raise TypeError(f"Expected data of type {self._dtype}, got {type(value)}")
        self._value = value
        self.emit("change", value)
        LOGGER.debug(
            f"Property '{self.name}' updated to {value} and 'change' event emitted."
        )



#!/usr/bin/env python
# ******************************************************************************
#
#  %W%  %G% CSS
#
#  "pyspec" Release %R%
#
#  Copyright (c) 2013,2014,2015,2016,2017,2018,2019,2020,2021
#  by Certified Scientific Software.
#  All rights reserved.
#
#  Permission is hereby granted, free of charge, to any person obtaining a
#  copy of this software ("pyspec") and associated documentation files (the
#  "Software"), to deal in the Software without restriction, including
#  without limitation the rights to use, copy, modify, merge, publish,
#  distribute, sublicense, and/or sell copies of the Software, and to
#  permit persons to whom the Software is furnished to do so, subject to
#  the following conditions:
#
#  The above copyright notice and this permission notice shall be included
#  in all copies or substantial portions of the Software.
#
#  Neither the name of the copyright holder nor the names of its contributors
#  may be used to endorse or promote products derived from this software
#  without specific prior written permission.
#
#     * The software is provided "as is", without warranty of any   *
#     * kind, express or implied, including but not limited to the  *
#     * warranties of merchantability, fitness for a particular     *
#     * purpose and noninfringement.  In no event shall the authors *
#     * or copyright holders be liable for any claim, damages or    *
#     * other liability, whether in an action of contract, tort     *
#     * or otherwise, arising from, out of or in connection with    *
#     * the software or the use of other dealings in the software.  *
#
# ******************************************************************************

"""

****************
filespec
****************

Description
****************
   This module offers an interface to files with data in the spec
   file format. The format specificiation can be consulted online
   at the certif.com website.

   A spec file normally consists of a series of header blocks and scan blocks.
   Data comes after scan blocks. Sometimes comment lines can be also
   be found in between blocks.
   These comment lines could contain for example the result of user
   calculations after a scan, or pure comments.  For practical purposes comment
   lines will be considered to belong to the preceding scan or header block.

.. _`certif.com`: http://www.certif.com/spec_help/scans.html

"""


import re
import numpy
import os
import time
import string


class FileSpecFormatUnknown(BaseException):
    pass


class FileSpec(list):
    """
    FileSpec class documentation
    """

    def __init__(self, filename):

        list.__init__(self)

        self._filename = filename
        self.origfilename = None
        self._headers = []
        self.lastpos = 0

        self.inheader = False

        self.filestat = os.stat(self._filename)
        self.st_size = 0

        # dictionary to hold references (by scan number) to the scanlist
        self.scans = {}

        self._indexscans()

        if len(self.scans) == 0:
            raise FileSpecFormatUnknown("No scans found in file %s" % self._filename)

    @property
    def abspath(self):
        return self.absolutePath()

    def absolutePath(self):
        return os.path.abspath(self._filename)

    def update(self):

        currstat = os.stat(self._filename)

        if currstat.st_size > self.st_size:
            self.st_size = currstat.st_size
            self._indexscans()
            modified = True
        else:
            modified = False

        return modified

    @property
    def filename(self):
        return self.getFileName()

    def getFileName(self):
        return self._filename

    def getScanByNumber(self, scanno, scanorder=0):
        if scanno in self.scans:
            if scanorder >= 0 and scanorder < len(self.scans[scanno]):
                scan = self.scans[scanno][scanorder]
                return scan
        else:
            return None

    get_scan_by_number = getScanByNumber

    @property
    def time_created(self):
        return self.getTimeCreated()

    def getTimeCreated(self):
        if self._headers:
            return self._headers[0].getDate()

    @property
    def user(self):
        return self.getUser()

    def getUser(self):
        if self._headers:
            return self._headers[0].getUser()

    @property
    def spec(self):
        return self.getSpec()

    def getSpec(self):
        if self._headers:
            return self._headers[0].getSpec()

    @property
    def time_modified(self):
        return self.getTimeModified()

    def getTimeModified(self):
        if not self._filename:
            return None

        mtime = os.stat(self._filename).st_mtime
        return time.asctime(time.localtime(mtime))

    @property
    def length(self):
        return self.getNumberScans()

    def getNumberScans(self):
        return len(self.scans)

    @property
    def headers(self):
        return self.getNumberHeaders()

    def getNumberHeaders(self):
        return len(self._headers)

    @property
    def info(self):
        return self.getInfo()

    def getInfo(self):
        """Returns user and application"""
        ctime = self.getTimeCreated()
        mtime = self.getTimeModified()
        user = self.getUser()
        spec = self.getSpec()
        return [ctime, mtime, user, spec]

    def _indexscans(self):

        self.fd = open(self._filename, "r")

        if len(self) > 0:
            fb = self[-1]
            self.fd.seek(self.lastpos)
        else:
            fb = None

        data = self.fd.read()

        for lineno, sline in enumerate(data.split("\n#")):

            if not sline.strip():
                continue

            if sline[0] in ["S", "F", "E"]:

                btype = sline[0]

                if btype in ["F", "E"]:
                    # block not followed by space is not a block. ignore line
                    if sline[1] != " ":
                        continue

                blockstart = self.lastpos
                blockline = lineno

                if fb is not None:
                    fb.end()

                if btype in ("F", "E"):
                    if btype == "F":
                        self.origfilename = sline[2:].strip()
                    fb = Header(blockstart, blockline)
                    self.inheader = True
                    self._headers.append(fb)
                elif btype == "S":
                    fb = Scan(blockstart, blockline)
                    fb.addSLine(sline[2:])
                    self.inheader = False
                    self.append(fb)
                    fb._setScanIndex(len(self))
                    if len(self._headers):
                        # Assign last added header to current scan
                        fb._setFileHeader(self._headers[-1])

                if self.origfilename:
                    fb.setFileName(self.origfilename)

            if sline and fb:
                fb.addLine(sline)

            self.lastpos = self.fd.tell()

        # register last block
        if fb is not None:
            fb.end()

        # correct the scan order if necessary
        # assign number in file

        self.scans = {}
        scanidx = 0

        for scan in self:
            scanno = scan.getNumber()
            if scanno not in self.scans:
                self.scans[scanno] = []

            self.scans[scanno].append(scan)
            scan._setOrder(len(self.scans[scanno]) - 1)
            scan._setNumberInFile(scanidx)
            scanidx += 1

        self.fd.close()


class FileBlock:

    respecuser = re.compile(r"(?P<spec>.*?)\s+User\s+=\s+(?P<user>.*?)$")

    def __init__(self, start, firstline):

        self.start = start
        self.firstline = firstline
        self.lines = []
        self._filename = None
        self._contains_error = False
        self._error_messages = []
        self._id = ""

        self.funcs = {
            "S": self.addSLine,
            "E": self.addEpochLine,
            "F": self.addFileLine,
            "D": self.addDateLine,
            "N": self.addColumnsLine,
            "L": self.addLabelLine,
            "O": self.addMotorLabelLine,
            "o": self.addMotorMneLine,
            "J": self.addCounterLabelLine,
            "j": self.addCounterMneLine,
            "U": self.addUserLine,
            "C": self.addCommentLine,
            "P": self.addMotorPositionLine,
            "T": self.addTimeLine,
            "G": self.addGeoLine,
            "Q": self.addQLine,
            "@": self.addExtraLine,
        }

        self.resetParsedData()

    def resetParsedData(self):
        # Default
        self.is_parsed = False

        self._data = []
        self._oneds = []

        # self._number = 0
        # self._command = ""

        self._count_time = 0
        self._filename = ""
        self._epoch = 0
        self._date = ""
        self._columns = 0
        self._labels = None
        self._motor_labels = []
        self._motor_mnes = []
        self._counter_labels = []
        self._counter_mnes = []
        self._motor_positions = []
        self._comment_lines = []
        self._user_lines = []
        self._geo_pars = []
        self._qvalue = 0
        self._extra_lines = []
        self._wrong_lines = []
        self._error_messages = []
        self._contains_error = False
        self._find_oned = True
        self.reading_mca = False

    def addLine(self, line):
        self.lines.append(line)

    def end(self):
        pass

    def parse(self):
        lineno = -1
        oned_idx = 0
        data_line = 0
        comp_line = 2  # The mca data is between 2 data counter lines.

        for line in self.lines:
            for sline in line.split("\n"):

                lineno += 1

                if not sline:
                    continue

                first_char = sline[0]

                if first_char in string.ascii_letters + "@":

                    widx = sline.find(" ")

                    metakey = sline[0]
                    metaval = sline[1:widx].strip()
                    content = sline[widx:].strip()

                    if metakey in self.funcs:
                        try:
                            self.funcs[metakey](content.strip(), metaval)
                        except:
                            self.wrongLine(
                                lineno, sline, "unknown line (%s) " % metakey
                            )
                    else:
                        self.wrongLine(
                            lineno, sline, "unknown header line (%s) " % metakey
                        )

                    if sline[0:2] == "@A":
                        if data_line == 1:
                            comp_line = 1  # The mca data is the first line.

                        sline = sline[2:]
                        self.reading_mca = True

                        if self._find_oned:
                            self._oneds.append(OneD())

                        self.tmpmca = McaData()
                        complete = self.tmpmca._addLine(sline)
                        if complete:
                            self._oneds[oned_idx].append(self.tmpmca)
                            self.reading_mca = False
                            oned_idx += 1
                else:
                    data_line += 1
                    if self.reading_mca:
                        complete = self.tmpmca._addLine(sline)
                        if complete:
                            self._oneds[oned_idx].append(self.tmpmca)
                            self.reading_mca = False
                            oned_idx += 1
                    else:
                        oned_idx = 0
                        try:
                            try:
                                dataline = list(map(float, sline.strip().split()))
                            except BaseException:
                                self.wrongLine(lineno, sline, "wrong data line")
                                continue

                            if len(dataline) != self._columns:
                                self.wrongLine(lineno, sline, "wrong number of columns")
                            else:
                                self._data.append(dataline)
                                if len(self._data) == comp_line:
                                    self._find_oned = False
                        except ValueError:
                            self.wrongLine(lineno, sline, "cannot parse line ")

        self.is_parsed = True

        self.finalizeParsing()

    def finalizeParsing(self):
        pass

    def wrongLine(self, lineno, sline, errmsg):
        self._wrong_lines.append([errmsg, sline])
        line = "%s (%s)" % (lineno + 1, self.firstline + lineno + 1)
        ermsg = "erroneous data / %s " % errmsg
        self._error_messages.append([self._id, line, ermsg])
        self._contains_error = True

    def setFileName(self, filename):
        self._filename = filename

    def addSLine(self, content, keyval=None):
        vals = content.split()
        self._number = int(vals[0])
        self._id = self._number
        self._command = " ".join(vals[1:])

    def addFileLine(self, content, keyval=None):
        self._filename = content

    def addEpochLine(self, content, keyval=None):
        self._epoch = int(content)

    def addDateLine(self, content, keyval=None):
        self._date = content

    def addColumnsLine(self, content, keyval=None):
        if not self._columns:
            self._columns = int(content)
        else:
            pass

    def addLabelLine(self, content, keyval=None):
        if self._labels is None:
            self._labels = re.split(r"\s\s+", content)
            if not self._columns:
                # cope with no N line. get nb columns from _labels
                self._columns = len(self._labels)
        else:
            pass

    def addMotorLabelLine(self, content, keyval=None):
        # Beware of double spacing
        self._motor_labels.extend(re.split(r"\s\s+", content))

    def addMotorMneLine(self, content, keyval=None):
        self._motor_mnes.extend(re.split(r"\s", content))

    def addCounterLabelLine(self, content, keyval=None):
        # Beware of double spacing
        self._counter_labels.extend(re.split(r"\s\s+", content))

    def addCounterMneLine(self, content, keyval=None):
        # Beware of double spacing
        self._counter_mnes.extend(re.split(r"\s", content))

    def addMotorPositionLine(self, content, keyval=None):
        self._motor_positions.extend(content.split(" "))

    def addUserLine(self, content, keyval=None):
        self._user_lines.append(content)

    def addCommentLine(self, content, keyval=None):
        self._comment_lines.append(content)

    def addTimeLine(self, content, keyval=None):
        parts = content.split()
        if len(parts) > 1:
            units = re.sub(r"[\)\(]", "", parts[1])
            self._count_time = [parts[0], units]
        else:
            self._count_time = [content, ""]

    def addGeoLine(self, content, keyval=None):
        self._geo_pars.append(content.split())

    def addQLine(self, content, keyval=None):
        self._qvalue = content

    def addExtraLine(self, content, keyval=None):
        self._extra_lines.append([keyval, content])

    @property
    def date(self):
        return self.getDate()

    def getDate(self):
        """
        Returns the date when the scan was started
        """
        if not self.is_parsed:
            self.parse()
        return self._date

    @property
    def user_spec(self):
        return self.getUserSpec()

    def getUserSpec(self):

        comments = self._comment_lines

        if comments:
            for line in comments:
                mat = self.respecuser.search(line)
                if mat:
                    return [mat.group("user"), mat.group("spec")]

        return [None, None]

    @property
    def spec(self):
        return self.getSpec()

    def getSpec(self):
        """
        Returns the name of the spec application
        from which the file was created
        """
        if not self.is_parsed:
            self.parse()

        return self.getUserSpec()[1]

    @property
    def user(self):
        return self.getUser()

    def getUser(self):
        """
        Returns the name of the unix user that created the file
        """
        if not self.is_parsed:
            self.parse()

        return self.getUserSpec()[0]


class Header(FileBlock):
    """
    Class representing a file header.
    """

    def __init__(self, start, firstline):
        FileBlock.__init__(self, start, firstline)

    def end(self):
        self.parse()


class Scan(FileBlock):
    """
    Scan class documentation
    """

    def __init__(self, start, firstline):
        FileBlock.__init__(self, start, firstline)
        self._fileheader = None
        self._numberinfile = -1
        self._order = 1

    def end(self):
        self.resetParsedData()

    def finalizeParsing(self):

        # prepare motor positions
        labels = self.getMotorNames()
        poss = self._motor_positions
        poserr = False
        self.motor_positions_list = None

        if not labels:
            ermsg = "no motor names"
            self._error_messages.append([self._id, "", ermsg])
            self._contains_error = True
            poserr = True

        elif len(labels) != len(poss):
            ermsg = "number of motor labels and positions are different"
            self._error_messages.append([self._id, "", ermsg])
            self._contains_error = True
            poserr = True

        if not poserr:
            self.motor_positions_list = list(zip(labels, poss))

    def _setFileHeader(self, header):
        self._fileheader = header

    def _setScanIndex(self, idx):
        self._index = idx

    def getScanIndex(self):
        """
        Returns the position of the scan in the file
        """
        return self._index

    @property
    def number(self):
        return self.getNumber()

    def getNumber(self):
        """
        Returns scan number as it appeared in spec.
        Remember that it could be that more than one scan
        in the file will have the same number.
        The scan index is the position of the scan in the file.
        The scan number is the number given by spec
        to the scan at the time it was executed.
        """
        return self._number

    @property
    def order(self):
        return self.getOrder()

    def getOrder(self):
        """
        Returns scan order for the scan. The combination
        of scan number/scan order should be unique for a scan.
        in a file.  The first scan with a certain number in
        the file will have order 1.  If another scan in the file
        uses the same number, it will be associated with order 2 and so on.
        """
        return self._order

    def _setNumberInFile(self, number):
        self._numberinfile = number

    def getNumberInFile(self):
        return self._numberinfile

    @property
    def nb_points(self):
        return self.getLines()

    def getLines(self):
        """
        Returns number of data lines
        """
        return len(self._data)

    @property
    def nb_columns(self):
        return self.getColumns()

    def getColumns(self):
        """
        Returns number of columns from scan header
        """
        if not self.is_parsed:
            self.parse()
        return self._columns

    @property
    def labels(self):
        return self.getLabels()

    def getLabels(self):
        """
        Returns the labels for the data columns
        """
        if not self.is_parsed:
            self.parse()
        return self._labels

    @property
    def command(self):
        return self.getCommand()

    def getCommand(self):
        """
        Returns a string containing the command that
        was run in spec to start the scan
        """
        return self._command

    @property
    def motor_names(self):
        return self.getMotorNames()

    def getMotorNames(self):
        """
        Returns a list with motor names
        """
        if not self.is_parsed:
            self.parse()

        if self._motor_labels:
            return self._motor_labels
        elif self._fileheader and self._fileheader._motor_labels:
            return self._fileheader._motor_labels
        else:
            return None

    @property
    def motors(self):
        return self.getMotorMnemonics()

    def getMotorMnemonics(self):
        """
        Returns a list with motor mnemonics. Motor mnemonics are saved
        in files only since spec version 6.0.10
        """
        if not self.is_parsed:
            self.parse()

        if self._motor_mnes:
            return self._motor_mnes
        elif self._fileheader and self._fileheader._motor_mnes:
            return self._fileheader._motor_mnes
        else:
            return None

    @property
    def counter_names(self):
        return self.getCounterNames()

    def getCounterNames(self):
        """
        Returns a list with counter names. Counter names are saved
        in files only since spec version 6.0.10
        """
        if not self.is_parsed:
            self.parse()

        if self._counter_labels:
            return self._counter_labels
        elif self._fileheader and self._fileheader._counter_labels:
            return self._fileheader._counter_labels
        else:
            return None

    @property
    def counters(self):
        return self.getCounterMnemonics()

    def getCounterMnemonics(self):
        """
        Returns a list with counter mnemonics. Counter mnemonics are saved
        in files only since spec version 6.0.10
        """
        if not self.is_parsed:
            self.parse()

        if self._counter_mnes:
            return self._counter_mnes
        elif self._fileheader and self._fileheader._counter_mnes:
            return self._fileheader._counter_mnes
        else:
            return None

    @property
    def motor_positions(self):
        return self.getMotorPositions()

    def getMotorPositions(self):
        """
        Returns a dictionary with motor names and positions.
        These are the positions of the motors when the scan was started
        """
        if not self.is_parsed:
            self.parse()

        return self.motor_positions_list

    @property
    def user(self):
        return self.getUser()

    def getUser(self):
        if self._fileheader:
            return self._fileheader.getUser()

    @property
    def spec(self):
        return self.getSpec()

    def getSpec(self):
        if self._fileheader:
            return self._fileheader.getSpec()

    @property
    def date(self):
        return self.getDate()

    def getDate(self):
        """
        Returns the date when the scan was started
        """
        if not self.is_parsed:
            self.parse()
        return self._date

    @property
    def file_date(self):
        return self.getFileDate()

    def getFileDate(self):
        """
        Returns the date when the file was created
        """
        if self._fileheader:
            return self._fileheader._date
        else:
            return None

    @property
    def source(self):
        return self.getSource()

    def getSource(self):
        """
        Returns the path of the file as it appears in the file header
        """
        if self._fileheader:
            if self._fileheader._filename:
                return self._fileheader._filename

        return ""

    @property
    def geometry(self):
        return self.getGeometry()

    def getGeometry(self):
        """
        Returns geometry values as saved in the file.
        Check the spec documentation for the meaning of these values
        """
        if not self.is_parsed:
            self.parse()
        return [" ".join(line) for line in self._geo_pars]

    @property
    def hkl(self):
        return self.getHKL()

    def getHKL(self):
        """
        Returns a list with HKL values at the beginning of the scan
        """
        if not self.is_parsed:
            self.parse()
        return self._qvalue

    @property
    def file_epoch(self):
        return self.getFileEpoch()

    def getFileEpoch(self):
        """
        Returns the epoch of the file creation. It is possible to find
        the absolute epoch for any scan time by adding
        the file epoch with the value in the Epoch column of the scan
        """
        if self._fileheader:
            return self._fileheader._epoch
        else:
            return None

    @property
    def count_time(self):
        return self.getCountTime()

    def getCountTime(self):
        """
        Returns a list with two values: counting time and units
        if time units cannot be found in file the units value is left empty
        """
        if not self.is_parsed:
            self.parse()
        return self._count_time

    @property
    def comments(self):
        return self.getComments()

    def getComments(self):
        """
        Returns comments in the scan.
        Aborted termination can be found in this way
        """
        if not self.is_parsed:
            self.parse()
        return self._comment_lines

    @property
    def user_lines(self):
        return self.getUserLines()

    def getUserLines(self):
        if not self.is_parsed:
            self.parse()
        return self._user_lines

    @property
    def extra_lines(self):
        return self.getExtraLines()

    def getExtra(self):
        """
        Returns extra lines starting with "@" character.
        These are normally lines related with MCA data
        """
        if not self.is_parsed:
            self.parse()
        return self.getExtraLines()

    def getExtraLines(self):
        if not self.is_parsed:
            self.parse()
        return [" ".join(line) for line in self._extra_lines]

    @property
    def metadata(self):
        return self.getMeta()

    def getMeta(self):
        """
        Returns a dictionary with the most relevant metdata information
        """
        if not self.is_parsed:
            self.parse()

        meta = {
            "spec": "",
            "user": "",
            "source": "",
            "HKL": "",
            "date": "",
            "scanno": "",
            "motors": None,
            "comments": None,
            "errors": None,
        }

        # spec and user. In fileheader comment line
        meta["spec"] = self.getSpec()
        meta["user"] = self.getUser()
        meta["source"] = self.getSource()
        meta["HKL"] = self.getHKL()
        meta["date"] = self.getDate()
        meta["scanno"] = self.getNumber()
        meta["motors"] = self.getMotorPositions()
        meta["motornames"] = self.getMotorNames()
        meta["comments"] = self.getComments()
        meta["order"] = self.getOrder()
        meta["noinfile"] = self.getNumberInFile()
        meta["points"] = self.getLines()
        meta["columns"] = self.getColumns()
        meta["userlines"] = self.getUserLines()
        meta["geo"] = self.getGeometry()
        meta["extra"] = self.getExtra()

        motmnes = self.getMotorMnemonics()
        if motmnes:
            meta["motormnes"] = self.getMotorMnemonics()

        if self._contains_error:
            meta["errors"] = self._error_messages

        return meta

    @property
    def data(self):
        return self.getData()

    def getData(self):
        """
        Returns a numpy array with all data in the scan
        """
        if not self.is_parsed:
            self.parse()

        if self._data:
            return numpy.array(self._data, dtype=float)
        else:
            return numpy.empty((0, self._columns))

    @property
    def nb_mcas(self):
        return self.getNumberMcas()

    def getNumberMcas(self):
        """
        Returns the number of mcas in the scan
        """
        if not self.is_parsed:
            self.parse()
        return sum(map(len, self._oneds))

    @property
    def mcas(self):
        return self.getMcas()

    def getMcas(self):
        """
        Returns a list of 1D numpy arrays in the scan,
        each of them being a spectrum from a 1D detector
        """
        if not self.is_parsed:
            self.parse()
        result = []
        for mcas in self._oneds:
            result += mcas
        return result

    @property
    def nb_oneds(self):
        return self.getNumberOneD()

    def getNumberOneD(self):
        """
        Returns the number of OneD channels in the scan.
        """
        if not self.is_parsed:
            self.parse()
        return len(self._oneds)

    def getOneD(self, idx):
        if not self.is_parsed:
            self.parse()
        return self._oneds[idx]

    get_mca = getOneD
    get_oned = getOneD

    def _setOrder(self, order):
        self._order = order

    def __str__(self):
        if not self.is_parsed:
            self.parse()
        if self._order > 1:
            return "%s.%s %s" % (self._number, self._order, self._command)
        else:
            return "%s %s" % (self._number, self._command)

    def save(self, outfile, format="spec", append=False, columns=None, mcas=False):
        """scan.save method produces a simple output meant to export scan
        data to format readable by excel and other programs
        """

        data = self.getData()
        meta = {}

        meta["command"] = self.getCommand()
        meta["number"] = self.getNumber()
        meta["columns"] = data.shape[1]

        if append:
            ofd = open(outfile, "a")
        else:
            ofd = open(outfile, "w")

        if format == "tabs":
            labsep = "\t"
            datsep = "\t"
            first = ""
        elif format == "csv":
            labsep = ","
            datsep = ","
            first = ""
        elif format == "spec":
            labsep = "  "
            datsep = " "
            first = (
                """
#S %(number)s %(command)s
#N %(columns)s
#L """
                % meta
            )

        ofd.write(first + labsep.join(self.getLabels()) + "\n")
        for row in range(data.shape[0]):
            outline = datsep.join(["%.12g" % val for val in data[row]]) + "\n"
            ofd.write(outline)
        ofd.write("\n")


class McaData:
    """
    The class MCA data represents 1D data
    """

    def __init__(self):
        self._data = []
        self._calib = None

    @property
    def calib(self):
        return self.getCalib()

    def getCalib(self):
        return self.calib

    @calib.setter
    def calib(self, calib):
        return self.setCalib(calib)

    def setCalib(self, calib):
        self._calib = calib

    @property
    def data(self):
        return self.getData()

    def getData(self, calibrated=False):
        channels = list(range(len(self._data)))

        if calibrated and self._calib:
            a, b, c = self._calib
            indexes = [(a + b * x + c * x**2) for x in channels]
        else:
            indexes = channels

        if self._data:
            return numpy.array([indexes, self._data], dtype=float).transpose()
        else:
            return numpy.empty((0, 1))

    def __len__(self):
        return len(self._data)

    def _addLine(self, line):
        if line.strip()[-1] == "\\":
            dataline = line.strip()[:-1]
            complete = False
        else:
            dataline = line
            complete = True

        self._data.extend(list(map(float, dataline.split())))
        return complete


class OneD(list):
    """
    The class OneD is for the one dimension channels.
    It has a list of McaData objects.
    """

    @property
    def data(self):
        return self.getData()

    def getData(self):
        data = []
        for mcadata in self:
            raw_data = mcadata.getData().transpose()[1].tolist()
            data.append(raw_data)
        return numpy.array(data, dtype=float)


if __name__ == "__main__":
    import sys

    t0 = time.time()

    fs = FileSpec(sys.argv[1])

    print("Time to open file was %s " % (time.time() - t0))
    print("  number of scans: %s" % len(fs))

    print("  file created by user %s" % fs.user)

    for scan in fs:
        print("Scan %d /  " % scan.number)
        print("   count time is %s /  " % str(scan.count_time))
        print("   number of points in scan: %d" % scan.nb_points)
        print("   number of mcas in scan: %d" % scan.nb_mcas)
        for mca in scan.mcas:
            print("     - mca / shape: %s" % str(mca.data.shape))



from contextlib import contextmanager
from email import contentmanager
from typing import Any, Callable, Literal, Mapping, Union, Iterable, TypeVar, overload, Dict, Tuple, Union, Optional
from pyee.asyncio import AsyncIOEventEmitter
import re

AssociativeArrayElement = Union[float, int, str]
AssociativeArrayKey = Union[
    AssociativeArrayElement, Tuple[AssociativeArrayElement, AssociativeArrayElement]
]


T = TypeVar("T")


def by_two(iterable: Iterable[T]) -> Iterable[tuple[T, T]]:
    """Yield successive pairs from an iterable."""
    a = iter(iterable)
    return zip(a, a)


def try_cast(value: str) -> AssociativeArrayElement:
    try:
        v = float(value)
        if v.is_integer():
            return int(v)
        return v
    except ValueError:
        return value


class AssociativeArray:
    """
    Represents a SPEC associative array, which is a collection of key-value pairs.

    See https://certif.com/spec_manual/ref_2_3_4_1.html for more details on how spec handles them.

    Ultimately, the associative array is a mapping from one or two keys to a value.

    Example usage:
    .. code-block:: python

        x = AssociativeArray()
        x["one"] = "now"
        x["three"] = "the"
        x["three", "0"] = "time"
        x["two"] = "is"

    All keys are stored as strings on spec with decimal numbers being expanded to something similar to ".5g" format.
    """

    KEY_SEPARATOR = "\x1c"
    ITEM_SEPARATOR = "\000"

    data: Dict[str, Optional[AssociativeArrayElement]]

    def __init__(self) -> None:
        super().__init__()
        self.data = {}

    def __getitem__(
        self,
        key: AssociativeArrayKey,
    ) -> AssociativeArrayElement:
        value = self.data.get(self.compose_key(key))
        if value is None:
            raise KeyError(f"Key {key} has been deleted from associative array")
        return value

    def __setitem__(
        self,
        key: AssociativeArrayKey,
        value: Optional[AssociativeArrayElement],
        /,
    ) -> None:
        self.data[self.compose_key(key)] = value

    def __delitem__(
        self,
        key: AssociativeArrayKey,
    ) -> None:
        del self.data[self.compose_key(key)]

    @classmethod
    def compose_key(cls, key: AssociativeArrayKey) -> str:
        if isinstance(key, tuple):
            key1, key2 = key
            if key2 == "":
                return cls.stringify_key(key1)
            return (
                f"{cls.stringify_key(key1)}{cls.KEY_SEPARATOR}{cls.stringify_key(key2)}"
            )
        return cls.stringify_key(key)

    @staticmethod
    def stringify_key(key: Union[float, int, str]) -> str:
        if isinstance(key, (float, int)):
            return f"{key:.5g}"
        return str(key)

    def serialize(self) -> bytes:
        key_value_list = []
        for key, value in self.data.items():
            key_value_list.append(key)
            key_value_list.append(value)

        return (
            self.ITEM_SEPARATOR.join(str(item) for item in key_value_list)
            + self.ITEM_SEPARATOR
        ).encode("utf-8")

    @classmethod
    def deserialize(cls, data: bytes, deleted: bool = False) -> "AssociativeArray":
        instance = cls()
        data_str = data.decode("utf-8")
        # Not sure why spec is sending an extra separator.
        for key, value in by_two(data_str.split(cls.ITEM_SEPARATOR)[:-1]):
            if deleted:
                instance.data[key] = None
            else:
                instance.data[key] = try_cast(value)
        return instance

    def __str__(self) -> str:
        items = list(self.data.items())
        if len(items) > 5:
            display_items = items[:5]
            display_str = "{" + ", ".join(
                f"{self.to_display_key(key)}: {value}" for key, value in display_items
            )
            display_str += ", ... }"
        else:
            display_str = (
                "{"
                + ", ".join(
                    f"{self.to_display_key(key)}: {value}" for key, value in items
                )
                + "}"
            )
        return display_str

    @staticmethod
    def to_display_key(key: str) -> str:
        if AssociativeArray.KEY_SEPARATOR not in key:
            key1 = key
            key2 = ""
        else:
            key1, key2 = key.split(AssociativeArray.KEY_SEPARATOR)
        return f"[{key1}][{key2}]" if key2 else f"[{key1}]"

    def update(self, other: "AssociativeArray") -> None:
        """
        Update the associative array with another associative array.

        Args:
            other (AssociativeArray): The other associative array to update from.
        """
        self.data.update(other.data)


# This is just:
# (not-brackets)* "[ .? ]" + maybe "[ .? ]"
# ?: is is used to form non-capturing groups for the parts of the regex we don't care about.
KEY_PATTERN = re.compile(r"^[^\[\]]*(?:\[(.*?)\])(?:\[(.*?)\])?$")


def get_associative_array_key(prop_name: str) -> Optional[Tuple[str, str]]:
    """
    Parse the property name to get the keys indexing into an associative array.

    Example:
    .. code-block:: none

        "../x" -> None
        "../x[1]" -> ("1", "")
        "../x[1][2]" -> ("1", "2")


    Args:
        prop_name (str): The property name to parse.
    """

    if not prop_name.endswith("]"):
        return None

    match = KEY_PATTERN.search(prop_name)
    if match is not None:
        groups = match.groups()
        # The groups from parsing are structured like:
        # ("../x", "key1",  "key2" or None)
        return groups[0], groups[1] or ""
    return None


def unpack_associative_array_element(property_name: str, value):
    """
    Given a property name and a value, if the property name indicates that this is an element of an associative array, extract the value of the element from the associative array.

    Example:
    .. code-block:: none

        property_name = "x[1]"
        value = AssociativeArray({("1", ""): "one"})
        returns "one"

        property_name = "x[1][2]"
        value = AssociativeArray({("1", "2"): "one two"})
        returns "one two"

        property_name = "x"
        value = ...
        returns value

    Args:
        property_name (str): The property name to parse for the associative array keys.
        value: The value to extract the element from if the property name indicates this is an element

    Returns:
        The extracted element value if the property name indicates this is an element of an associative array, otherwise the original value.
    """
    if (
        isinstance(value, AssociativeArray)
        and (key := get_associative_array_key(property_name)) is not None
    ):
        return value[key]
    return value


def pack_associative_array_element(property_name: str, value):
    """
    Given a property name and a value, if the property name indicates that this is an element of an associative array, pack the value into an associative array with the appropriate key.

    Example:
    .. code-block:: none

        property_name = "x[1]"
        value = "one"
        returns AssociativeArray({("1", ""): "one"})

        property_name = "x[1][2]"
        value = "one two"
        returns AssociativeArray({("1", "2"): "one two"})
        property_name = "x"
        value = ...
        returns value
    """
    if (key := get_associative_array_key(property_name)) is not None:
        arr = AssociativeArray()
        arr[key] = value
        return arr
    return value



import asyncio

import pytest
from conftest import HOST, PORT
from exceptiongroup import suppress

import pyspec._connection
import pyspec._connection.connection
from pyspec._connection import ClientConnection
from pyspec._connection.client_connection import RemoteException
from pyspec._connection.data import ErrorStr


@pytest.mark.asyncio
async def test_client_connect(server_process):
    async with ClientConnection(HOST, PORT) as client:
        assert client.host == HOST
        assert client.port == PORT


@pytest.mark.asyncio
async def test_command_execution(server_process):
    async with ClientConnection(HOST, PORT) as client:
        # Only works if server is in test_mode
        result = await client.remote_cmd("1+1")
        assert result == 2


@pytest.mark.asyncio
async def test_function_execution_success(server_process):
    async with ClientConnection(HOST, PORT) as client:
        result = await client.remote_func("sum", 1, 2)
        assert result == 3


@pytest.mark.asyncio
async def test_async_function_execution_success(server_process):
    async with ClientConnection(HOST, PORT) as client:
        result = await client.remote_func("async_sum", 1, 2)
        assert result == 3


@pytest.mark.asyncio
async def test_function_execution_not_found(server_process):
    async with ClientConnection(HOST, PORT) as client:
        with pytest.raises(RemoteException):
            result = await client.remote_func("doesnt-exist-on-server", 1, 2)


@pytest.mark.asyncio
async def test_property_read_write(server_process):
    async with ClientConnection(HOST, PORT) as client:
        await client.prop_set("foo", 123)
        value = await client.prop_get("foo")
        assert value == 123


@pytest.mark.asyncio
async def test_property_not_found(server_process):
    async with ClientConnection(HOST, PORT) as client:
        with pytest.raises(RemoteException):
            await client.prop_get("doesnt-exist-on-server")

        # Doesn't raise
        # These errors would show up under /error events,
        # since these are "fire-and-forget" operations
        await client.prop_set("doesnt-exist-on-server", 123)
        await client.prop_watch("doesnt-exist-on-server")
        await client.prop_unwatch("doesnt-exist-on-server")


@pytest.mark.asyncio
async def test_property_subscribe(server_process):
    async with ClientConnection(HOST, PORT) as client:
        updates = []
        client.on("property-change", lambda name, value: updates.append((name, value)))
        await client.prop_watch("ticker")
        await client.prop_set("ticker", 42)
        await asyncio.sleep(0.05)
        assert ("ticker", 42) in updates


@pytest.mark.asyncio
async def test_command_and_function_interrupt(server_process):
    async with ClientConnection(HOST, PORT) as client:
        # This is a placeholder; actual interruption logic depends on server implementation
        task = asyncio.create_task(client.remote_func("long_running_func"))
        await asyncio.sleep(0.05)
        await client.abort()
        with pytest.raises(RemoteException):
            await task


@pytest.mark.asyncio
async def test_command_and_function_timeout(server_process):
    async with ClientConnection(HOST, PORT) as client:
        # This is a placeholder; actual timeout logic depends on server implementation

        error_fut = asyncio.Future()

        def on_msg(msg: pyspec._connection.connection.Connection.Message):
            if isinstance(msg.data, ErrorStr):
                with suppress(asyncio.InvalidStateError):
                    error_fut.set_result(msg.data)

        client.on("message", on_msg)

        with suppress(asyncio.TimeoutError):
            await asyncio.wait_for(
                client.remote_func("long_running_func"),
                timeout=0.1,
            )

        task = asyncio.create_task(client.remote_func("long_running_func"))

        error_result = await asyncio.wait_for(error_fut, timeout=1)
        assert isinstance(error_result, ErrorStr)
        assert error_result == "Function execution aborted."



import pytest
import os
import pyspec
from pyspec import Client
import asyncio

from pyspec._connection.associative_array import AssociativeArray
from pyspec.client import Property

SERVER_PORT: int = int(os.environ.get("SPEC_SERVER_PORT", -1))
pytestmark = pytest.mark.skipif(
    SERVER_PORT == -1,
    reason="Requires SPEC_SERVER_PORT environment variable to be set to the port of a running spec server",
)


@pytest.mark.asyncio
async def test_connect():
    from pyspec import Client

    async with Client("localhost", SERVER_PORT) as client:
        assert client._connection is not None


@pytest.mark.asyncio
async def test_var_read_write():
    async with Client("localhost", SERVER_PORT) as client:
        test_var = client.var("test_var")
        await test_var.set(123)
        value = await test_var.get()
        assert value == 123


@pytest.mark.asyncio
async def test_var_subscribe():
    async with Client("localhost", SERVER_PORT) as client:
        async with client.var("test_var").subscribed() as test_var:
            await test_var.set(10)

            async with test_var.wait_for(123):
                await client.exec("test_var = 123")

            assert await test_var.get() == 123


@pytest.mark.asyncio
async def test_var_read_doesnt_exist():
    async with Client("localhost", SERVER_PORT) as client:
        with pytest.raises(pyspec.RemoteException):
            await client.var("doesnt_exist").get()


@pytest.mark.asyncio
async def test_output():
    async with Client("localhost", SERVER_PORT) as client:
        async with client.output("tty").capture() as lines:
            await client.exec('print("Hello, world!")')
        assert lines[-1] == "Hello, world!\n"


@pytest.mark.asyncio
async def test_status():
    async with Client("localhost", SERVER_PORT) as client:
        status = client.status()
        async with status:
            assert await status.ready.get()
            assert not await status.shell.get()
            assert await status.simulate.get()


@pytest.mark.asyncio
async def test_motor():
    async with Client("localhost", SERVER_PORT) as client:
        s0v = client.motor("s0v")
        async with s0v:
            await asyncio.sleep(1)
            current_position = await s0v.position.get()
            await s0v.move(-current_position)
        assert await s0v.position.get() == -current_position


@pytest.mark.asyncio
async def test_sync_motors():
    """
    Here is a communication log from SPEC when the server is running with REAL motors.
    This is the expected communication pattern for this test.

    Simulation mode behaves differently and sends the position event with the old position.

    sR      1    0.00  2 HELLO
    sW      1    0.00  2 HELLO_REPLY      spec                            spec
    sR      0    0.05  2 REGISTER         motor/xrfx/move_done
    sW      0   56.79  2 EVENT            motor/xrfx/dial_position        -24.40375
    sR      0    0.02  2 REGISTER         motor/xrfx/position
    sW      0    0.00  2 EVENT            motor/xrfx/move_done            0
    sW      0    0.00  2 EVENT            motor/xrfx/position             25.000001373291
    sR      0    0.04  2 REGISTER         motor/xrfz/move_done
    sW      0    0.00  2 EVENT            motor/xrfz/move_done            0
    sR      0    0.04  2 REGISTER         motor/xrfz/position
    sW      0    0.01  2 EVENT            motor/xrfz/position             326.000000610352
    sR      2    0.15  2 CHAN_READ        motor/xrfx/position
    sW      2    0.01  2 REPLY            motor/xrfx/position             25.000001373291
    sR      0  113.65  2 CHAN_SEND        motor/../prestart_all
    sR      0    0.36  2 CHAN_SEND        motor/xrfx/start_one            26
    sR      0    0.12  2 CHAN_SEND        motor/xrfz/start_one            325
    sR      0    0.08  2 CHAN_SEND        motor/../start_all
    sW      0  122.03  2 EVENT            motor/xrfx/move_done            1
    sW      0    1.62  2 EVENT            motor/xrfz/move_done            1
    sW      0    0.02  2 EVENT            motor/xrfx/position             25.000751373291
    sW      0    0.01  2 EVENT            motor/xrfz/position             325.999063110352
    sW      0    0.09  2 EVENT            motor/xrfx/position             26.000001373291
    sW      0    0.00  2 EVENT            motor/xrfx/move_done            0
    sW      0    0.05  2 EVENT            motor/xrfz/position             324.987875610352
    sW      0    0.03  2 EVENT            motor/xrfz/position             325.000000610352
    sW      0    0.02  2 EVENT            motor/xrfz/move_done            0
    """

    async with Client("localhost", SERVER_PORT) as client:
        xrfx = client.motor("xrfx")
        xrfz = client.motor("xrfz")

        xrfx_target = 26
        xrfz_target = 325

        async with xrfx, xrfz:
            with pytest.raises(
                RuntimeError, match="Cannot prepare move when not synchronizing motors"
            ):
                xrfx.prepare_move(xrfx_target)
                xrfz.prepare_move(xrfz_target)

            async with client.synchronized_motors():
                with pytest.raises(
                    RuntimeError, match="Cannot start move when synchronizing motors"
                ):
                    await xrfx.move(xrfx_target)
                    await xrfz.move(xrfz_target)

                xrfx.prepare_move(xrfx_target)
                xrfz.prepare_move(xrfz_target)

            # Note: There appears to be a bug in server mode simulation where the motor position
            # event doesn't actually get updated. So these tests might fail in with SIMULATION mode on.
            assert pytest.approx(xrfx_target, rel=1e-3) == await xrfx.position.get()
            assert pytest.approx(xrfz_target, rel=1e-3) == await xrfz.position.get()


@pytest.mark.asyncio
async def test_associative_array_index():
    async with Client("localhost", SERVER_PORT) as client:
        x = client.var("x[1]")
        await x.set("one")
        assert await x.get() == "one"
        await x.set("two")
        assert await x.get() == "two"


@pytest.mark.asyncio
async def test_associative_array_multi_index():
    async with Client("localhost", SERVER_PORT) as client:
        x = client.var("x[1][2]")
        await x.set("one two")
        assert await x.get() == "one two"
        await x.set("three four")
        assert await x.get() == "three four"


@pytest.mark.asyncio
async def test_associative_array():
    async with Client("localhost", SERVER_PORT) as client:
        x: Property[AssociativeArray] = client.var("x")
        xv = await x.get()
        assert isinstance(xv, AssociativeArray)

        xv[1, 2] = "three four"
        await x.set(xv)
        assert (await x.get())[1, 2] == "three four"



import h5py
import sys
import os


class h5crawler_error(Exception):
    pass


class h5crawler(object):
    def __init__(self, fileobj):
        if isinstance(fileobj, str):
            self.hfile = h5py.File(fileobj)
        elif isinstance(fileobj, h5py.File):
            self.hfile = fileobj
        else:
            self.hfile = None

    # search elements with key=value. returns full_path
    def search_key(self, search_key, obj=None, kyroot=None):
        self.search_wrapper(
            search_func=self._search_key_function, pars={"search_key": search_key}
        )
        return self.result

    def _search_key_function(self, pars, child, full_key, key):
        search_key = pars["search_key"]
        if key == search_key:
            return full_key
        return False

    # search datasets with 2d shape
    def search_2d(self):
        self.search_wrapper(search_func=self._search_2d_function, pars=None)
        return self.result

    def _search_2d_function(self, pars, child, full_key, key):
        if isinstance(child, h5py.Dataset):
            if len(child.shape) >= 2:
                return [full_key, child.shape]

    def search_wrapper(self, search_func, pars, obj=None, kyroot=None):

        if obj is None:
            obj = self.hfile
            self.result = []
            kyroot = "/"

        for ky in obj.keys():
            cky = os.path.join(kyroot, ky)
            child = self.hfile[cky]

            r = search_func(pars, child, full_key=cky, key=ky)

            if r:
                self.result.append(r)

            if not isinstance(child, h5py.Dataset):
                self.search_wrapper(search_func, pars, obj=child, kyroot=cky)

        return self.result


if __name__ == "__main__":
    cr = h5crawler(sys.argv[1])

    result = cr.search_2d()
    # result = cr.search_key('data')
    for r in result:
        print(r)





import asyncio
import logging
from contextlib import asynccontextmanager
from typing import Any, Awaitable, Callable, Literal, overload, Optional

from pyee.asyncio import AsyncIOEventEmitter

from .connection import Connection
from .data import DataType, ErrorStr
from .protocol import Command, Header


class ServerConnectionEventEmitter(AsyncIOEventEmitter):
    """
    Defines the typed events emitted by the ServerConnection.
    """

    @overload
    def emit(self, event: Literal["close"]) -> bool: ...
    @overload
    def on(self, event: Literal["close"], func: Callable[[], Any]) -> Callable[[], Any]:
        """
        Register an event listener for the 'close' event.

        A 'close' event is emitted whenever the client connection is closed.
        """

    @overload
    def emit(self, event: Literal["abort"]) -> bool: ...
    @overload
    def on(self, event: Literal["abort"], func: Callable[[], Any]) -> Callable[[], Any]:
        """
        Register an event listener for the 'abort' event.

        An 'abort' event is emitted whenever the client sends an ABORT command.
        The server should treat this like a "^C" from the keyboard with respect to any currently executing commands.
        """

    @overload
    def emit(self, event: Literal["hello"], sequence_number: int) -> bool: ...
    @overload
    def on(
        self, event: Literal["hello"], func: Callable[[int], Any]
    ) -> Callable[[int], Any]:
        """
        Register an event listener for the 'hello' event.

        A 'hello' event is emitted whenever the client sends a HELLO command.
        The server should respond with a HELLO_REPLY message.
        """

    @overload
    def emit(self, event: Literal["remote-cmd-no-return"], command: str) -> bool: ...
    @overload
    def on(
        self,
        event: Literal["remote-cmd-no-return"],
        func: Callable[[str], Any],
    ) -> Callable[[str], Any]:
        """
        Register an event listener for the 'remote-cmd-no-return' event.

        A 'remote-cmd-no-return' event is emitted whenever the client sends a command
        that does not expect a return value.
        """

    @overload
    def emit(
        self, event: Literal["remote-cmd"], sequence_number: int, command: str
    ) -> bool: ...
    @overload
    def on(
        self,
        event: Literal["remote-cmd"],
        func: Callable[[int, str], Awaitable[DataType]],
    ) -> Callable[[int, str], Awaitable[DataType]]:
        """
        Register an event listener for the 'remote-cmd' event.

        A 'remote-cmd' event is emitted whenever the client sends a command
        that expects a return value.
        """

    @overload
    def emit(
        self, event: Literal["remote-func-no-return"], function_call: str
    ) -> bool: ...
    @overload
    def on(
        self,
        event: Literal["remote-func-no-return"],
        func: Callable[[str], Any],
    ) -> Callable[[str], Any]:
        """
        Register an event listener for the 'remote-func-no-return' event.

        A 'remote-func-no-return' event is emitted whenever the client calls a function
        that does not expect a return value.

        A function call is represented as a string, e.g. "my_function(1, 'arg2')".
        """

    @overload
    def emit(
        self, event: Literal["remote-func"], sequence_number: int, function_call: str
    ) -> bool: ...
    @overload
    def on(
        self,
        event: Literal["remote-func"],
        func: Callable[[int, str], Awaitable[DataType]],
    ) -> Callable[[int, str], Awaitable[DataType]]:
        """
        Register an event listener for the 'remote-func' event.

        A 'remote-func' event is emitted whenever the client calls a function
        that expects a return value.

        A function call is represented as a string, e.g. "my_function(1, 'arg2')".
        """

    @overload
    def emit(
        self, event: Literal["property-set"], property_name: str, value: DataType
    ) -> bool: ...
    @overload
    def on(
        self,
        event: Literal["property-set"],
        func: Callable[[str, DataType], Any],
    ) -> Callable[[str, DataType], Any]:
        """
        Register an event listener for the 'property-set' event.

        A 'property-set' event is emitted whenever the client sets a property value on the server.
        """

    @overload
    def emit(
        self, event: Literal["property-get"], sequence_number: int, property_name: str
    ) -> bool: ...
    @overload
    def on(
        self,
        event: Literal["property-get"],
        func: Callable[[int, str], Any],
    ) -> Callable[[int, str], Any]:
        """
        Register an event listener for the 'property-get' event.

        A 'property-get' event is emitted whenever the client requests a property value from the server.
        """

    @overload
    def emit(self, event: Literal["property-watch"], property_name: str) -> bool: ...
    @overload
    def on(
        self,
        event: Literal["property-watch"],
        func: Callable[[str], Any],
    ) -> Callable[[str], Any]:
        """
        Register an event listener for the 'property-watch' event.

        A 'property-watch' event is emitted whenever the client registers to watch a property on the server.
        The server should start sending events to the client when the property value changes.
        """

    @overload
    def emit(self, event: Literal["property-unwatch"], property_name: str) -> bool: ...
    @overload
    def on(
        self,
        event: Literal["property-unwatch"],
        func: Callable[[str], Any],
    ) -> Callable[[str], Any]:
        """
        Register an event listener for the 'property-unwatch' event.

        A 'property-unwatch' event is emitted whenever the client unregisters from watching a property on the server.
        The server should stop sending events to the client when the property value changes.
        """

    @overload
    def on(
        self, event: Literal["message"], func: Callable[["ServerConnection.Message"], Any]
    ) -> Callable[["ServerConnection.Message"], Any]:
        """
        Register an event listener for the 'message' event.

        A 'message' event is emitted whenever a new message is received from the client.
        Consider using the more specific events defined in this class instead.
        """

    def emit(self, event: str, *args: Any) -> bool:  # type: ignore[override]
        return super().emit(event, *args)

    def on(self, event: str, func: Optional[Callable[..., Any]] = None):  # type: ignore[override]
        if func is None:
            return super().on(event)
        return super().on(event, func)


class ServerConnection(Connection, ServerConnectionEventEmitter):
    def __init__(self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter):
        """
        Initialize a server connection with the given stream reader and writer.

        Args:
            reader (asyncio.StreamReader): The stream reader for the connection.
            writer (asyncio.StreamWriter): The stream writer for the connection.
        """
        Connection.__init__(self, reader, writer)
        self.on("message", self._dispatch_typed_message_events)
        self.logger = logging.getLogger("pyspec.server").getChild(
            f"{self.host}:{self.port}"
        )

    def _dispatch_typed_message_events(self, msg: "ServerConnection.Message") -> None:
        """
        Given a received message, emit the appropriate typed event based on the message command.
        Validates the message data type where possible.

        Args:
            msg (ServerConnection.Message): The received message.
        """

        cmd = msg.header.command
        if cmd == Command.CLOSE:
            self.emit("close")
        elif cmd == Command.ABORT:
            self.emit("abort")
        elif cmd == Command.HELLO:
            self.emit("hello", msg.header.sequence_number)
        elif cmd == Command.CHAN_SEND:
            self.emit("property-set", msg.header.name, msg.data)
        elif cmd == Command.CHAN_READ:
            self.emit("property-get", msg.header.sequence_number, msg.header.name)
        elif cmd == Command.REGISTER:
            self.emit("property-watch", msg.header.name)
        elif cmd == Command.UNREGISTER:
            self.emit("property-unwatch", msg.header.name)
        elif cmd in (Command.CMD, Command.CMD_WITH_RETURN, Command.FUNC, Command.FUNC_WITH_RETURN):
            if not isinstance(msg.data, str):
                raise TypeError(f"Expected command data to be str, got {type(msg.data)}")
            if cmd == Command.CMD:
                self.emit("remote-cmd-no-return", msg.data)
            elif cmd == Command.CMD_WITH_RETURN:
                self.emit("remote-cmd", msg.header.sequence_number, msg.data)
            elif cmd == Command.FUNC:
                self.emit("remote-func-no-return", msg.data)
            elif cmd == Command.FUNC_WITH_RETURN:
                self.emit("remote-func", msg.header.sequence_number, msg.data)
        else:
            raise ValueError(f"Unknown command: {msg.header.command}")

    async def prop_send(self, property: str, value) -> None:
        """
        Sends a property value to the client.

        Args:
            property (str): The property name.
            value: The value to send.
        """
        await self._send(Header(Command.EVENT, name=property), data=value)

    async def serve_forever(self) -> None:
        """
        Runs the server connection, handling requests until the connection is closed.
        """
        if self._listener is None:
            raise RuntimeError("Connection is not started.")
        await self._listener

    async def reply(self, sequence_number: int, data: DataType) -> None:
        """
        Sends a reply to the client for a given sequence number.

        Args:
            sequence_number (int): The sequence number to reply to.
            data (DataType): The data to send in the reply.
        """
        await self._send(
            Header(Command.REPLY, sequence_number=sequence_number), data=data
        )

    async def reply_error(self, sequence_number: int, error_message: str) -> None:
        """
        Sends an error reply to the client for a given sequence number.

        Args:
            sequence_number (int): The sequence number to reply to.
            error_message (str): The error message to send.
        """
        await self._send(
            Header(Command.REPLY, sequence_number=sequence_number),
            data=ErrorStr(error_message),
        )

    @asynccontextmanager
    async def catch_reply_exceptions(self, sequence_number: int):
        """
        Context manager to catch and handle exceptions during reply handling.

        Args:
            sequence_number (int): The sequence number for the reply.
        """
        try:
            yield
        except Exception as e:
            self.logger.error(
                "Exception occurred while handling command (%s): `%s`",
                sequence_number,
                e,
            )
            await self.reply_error(sequence_number, str(e))

    async def hello_reply(self, sequence_number: int) -> None:
        """
        Sends a HELLO_REPLY message to the client for a given sequence number.

        Args:
            sequence_number (int): The sequence number to reply to.
        """
        await self._send(Header(Command.HELLO_REPLY, sequence_number=sequence_number))



from __future__ import annotations

import ast
import asyncio
import inspect
import logging
import threading
from collections import defaultdict
from contextlib import contextmanager
from typing import Any, Callable

from pyee.asyncio import AsyncIOEventEmitter

from pyspec._connection.data import DataType
from pyspec.server._remote_property import Property

from .._connection import ServerConnection
from ._remote_function import (
    SyncOrAsyncCallable,
    is_remote_function,
    parse_remote_function_string,
    remote_function_name,
)

LOGGER = logging.getLogger("pyspec.server")


class Singleton:
    _lock = threading.Lock()

    def __new__(cls, *args, **kwargs):
        with cls._lock:
            if not hasattr(cls, "_instance"):
                cls._instance = super(Singleton, cls).__new__(cls)
        return cls._instance

    @classmethod
    def dispose(cls):
        """
        Disposes of the singleton instance. This can be useful in testing scenarios
        where you want to reset the state of the server between tests.
        """
        with cls._lock:
            if hasattr(cls, "_instance"):
                del cls._instance
                LOGGER.debug(f"{cls.__name__} singleton instance disposed.")


class ServerException(Exception):
    def __init__(self, msg: str, *args: object) -> None:
        """
        Exception for server errors.

        Args:
            msg (str): The error message.
        """
        super().__init__(msg, *args)
        self.msg = msg

    def __str__(self) -> str:
        return self.msg


class Server(AsyncIOEventEmitter, Singleton):

    def __init__(
        self,
        host: str = "localhost",
        port: int | None = None,
        allow_remote_code_execution: bool = False,
    ):
        """
        Initialize the server.

        Args:
            host (str, optional): The hostname to bind the server to.
            port (int | None, optional): The port to bind the server to.
            allow_remote_code_execution (bool, optional): If True, enables test mode (arbitrary code execution allowed).
        """
        super().__init__()
        self.host = host
        self.port = port
        self._server: asyncio.Server | None = None
        self._property_listeners: dict[str, set[ServerConnection]] = defaultdict(set)
        self._abortable_tasks: set[asyncio.Task] = set()
        self._remote_functions = self._build_remote_function_table()
        self._remote_properties = self._register_properties()

        self._allow_remote_code_execution = allow_remote_code_execution

        if self._allow_remote_code_execution:
            LOGGER.warning(
                "Server is running in ALLOW_REMOTE_CODE_EXECUTION mode. Arbitrary code execution is allowed."
            )

    def _build_remote_function_table(self):
        """
        Builds the table of remote functions by inspecting the Server instance for
        methods decorated with @remote_function.

        This table is used to dispatch remote function calls from clients.
        Returns:
            dict: Dictionary mapping function names to callables.
        """
        remote_functions: dict[str, SyncOrAsyncCallable] = {}
        for attr_name, attr in inspect.getmembers(self):
            if is_remote_function(attr):
                LOGGER.debug("Registering remote function: `%s`", attr_name)
                remote_functions[remote_function_name(attr)] = attr
        return remote_functions

    def _register_properties(self):
        """
        Registers all Property attributes found on the Server instance.
        This method sets up listeners to broadcast property changes to connected clients.
        Returns:
            dict: Dictionary mapping property names to Property objects.
        """

        def make_broadcaster(property_name: str) -> Callable[[Any], asyncio.Task]:
            return lambda value: asyncio.create_task(
                self.broadcast_property(property_name, value)
            )

        remote_properties: dict[str, Property[Any]] = {}
        for attr_name, prop in inspect.getmembers(self):
            if isinstance(prop, Property):
                LOGGER.debug(
                    "Registering remote property: `%s` at `%s`", attr_name, prop.name
                )
                prop.on("change", make_broadcaster(prop.name))
                remote_properties[prop.name] = prop
        return remote_properties

    @contextmanager
    def abortable(self, task: asyncio.Task):
        """
        Context manager to register a task as an abortable task on the server.

        A task registered like this will be interrupted by ABORT commands from clients or by calling the server's abort() method.

        Args:
            task (asyncio.Task): The task to register as abortable.
        """
        try:
            self._abortable_tasks.add(task)
            yield
        finally:
            self._abortable_tasks.discard(task)

    async def execute_command(self, command: str) -> DataType:
        """
        WARNING: This method allows execution of arbitrary code.
        You should not allow this to be called in production environments.

        Executes an arbitrary command string in test mode only.

        Args:
            command (str): The command string to execute.
        Returns:
            DataType: The result of the executed command.
        Raises:
            PermissionError: If not in test mode.
        """
        if self._allow_remote_code_execution:
            return eval(command)
        else:
            try:
                return ast.literal_eval(command)
            except Exception:
                LOGGER.warning(
                    "Attempted to execute non-literal command in non-test mode: %s",
                    command,
                )
                raise PermissionError("Command execution is only allowed in test mode.")

    async def execute_function(self, function_call: str) -> DataType:
        """
        Attempts to execute a function call defined on the server.
        Refers to the table of remote functions built at server initialization.

        Args:
            function_call (str): The function call string, e.g. "func_name(arg1, arg2)".
        Returns:
            DataType: The result of the function call.
        """

        name, args = parse_remote_function_string(function_call)
        if name not in self._remote_functions:
            raise ValueError(f"Remote function '{name}' not found on server.")

        # These are already bound to self if necessary, since they came from
        # getattr(self, attr_name)
        func = self._remote_functions[name]
        try:
            result = func(*args)
            if asyncio.iscoroutine(result):
                task = asyncio.create_task(result)
                with self.abortable(task):
                    return await task
            return result
        except asyncio.CancelledError:
            LOGGER.warning("Function '%s' aborted.", name)
            raise RuntimeError("Function execution aborted.")

    async def broadcast_property(self, property_name: str, value: DataType) -> None:
        """
        Broadcasts a property value to all connected clients that are subscribed to updates for that property.

        Args:
            property_name (str): The name of the property to broadcast.
            value (DataType): The value to broadcast.
        """
        listening_clients = self._property_listeners.get(property_name, set())
        await asyncio.gather(
            *[client.prop_send(property_name, value) for client in listening_clients]
        )

    async def abort(self):
        """
        Aborts all abortable tasks on the server.
        See the abortable() context manager.
        """
        for task in list(self._abortable_tasks):
            task.cancel()
        self._abortable_tasks.clear()

    async def _on_client_connected(
        self, client_reader: asyncio.StreamReader, client_writer: asyncio.StreamWriter
    ):
        """
        Handles a new client connection. Sets up event handlers for all of the necessary
        messages that the server needs to handle to implement a "SPEC Server."

        Args:
            client_reader (asyncio.StreamReader): The stream reader for the client connection.
            client_writer (asyncio.StreamWriter): The stream writer for the client connection.
        """
        host, port = client_writer.get_extra_info("peername")
        logger = LOGGER.getChild(f"{host}:{port}")
        connection = ServerConnection(client_reader, client_writer)

        def on_close():
            client_writer.close()

        async def on_abort():
            await self.abort()

        async def on_hello(sequence_number: int):
            await connection.hello_reply(sequence_number)

        async def on_cmd_no_return(command: str):
            logger.info("Command: %s", command)
            return await self.execute_command(command)

        async def on_cmd(sequence_number: int, command: str):
            async with connection.catch_reply_exceptions(sequence_number):
                await connection.reply(sequence_number, await on_cmd_no_return(command))

        async def on_func_no_return(function_call: str):
            logger.info("Function call: %s", function_call)
            return await self.execute_function(function_call)

        async def on_func(sequence_number: int, function_call: str):
            async with connection.catch_reply_exceptions(sequence_number):
                await connection.reply(
                    sequence_number, await on_func_no_return(function_call)
                )

        async def on_property_get(sequence_number: int, property_name: str):
            logger.info("Property get: %s", property_name)
            async with connection.catch_reply_exceptions(sequence_number):
                if property_name not in self._remote_properties:
                    raise ServerException(
                        f"Property '{property_name}' not found on server."
                    )
                prop = self._remote_properties[property_name]
                await connection.reply(sequence_number, prop.get())

        async def on_property_set(property_name: str, value: DataType):
            logger.info("Property set: %s", property_name)
            logger.debug("Property value: %s", value)
            if property_name not in self._remote_properties:
                logger.error("Property '%s' not found on server.", property_name)
                return
            prop = self._remote_properties[property_name]
            prop.set(value)

        async def on_property_watch(property_name: str):
            logger.info("Subscribed to property: %s", property_name)
            self._property_listeners[property_name].add(connection)
            # Immediately send the current value of the property upon subscription
            if property_name in self._remote_properties:
                prop = self._remote_properties[property_name]
                await connection.prop_send(property_name, prop.get())

        async def on_property_unwatch(property_name: str):
            logger.info("Unsubscribed from property: %s", property_name)
            self._property_listeners[property_name].discard(connection)

        async with connection:
            connection.on("close", on_close)
            connection.on("abort", on_abort)
            connection.on("remote-cmd-no-return", on_cmd_no_return)
            connection.on("remote-cmd", on_cmd)
            connection.on("remote-func-no-return", on_func_no_return)
            connection.on("remote-func", on_func)
            connection.on("property-get", on_property_get)
            connection.on("property-set", on_property_set)
            connection.on("property-watch", on_property_watch)
            connection.on("property-unwatch", on_property_unwatch)
            connection.on("hello", on_hello)

            logger.info("Connected")

            await connection.serve_forever()

    async def __aenter__(self):
        """
        Async context manager entry. Starts the server.
        """
        self._server = await asyncio.start_server(
            self._on_client_connected, self.host, self.port, reuse_port=True
        )
        return self

    async def __aexit__(self, exc_type, exc, tb):
        """
        Async context manager exit. Shuts down the server and disposes the singleton instance.
        """
        if self._server:
            self._server.close()
            await self._server.wait_closed()
            self._server = None
            LOGGER.info("Server shut down.")
        # Dispose of the singleton instance to allow for fresh instances in testing or future runs
        self.dispose()

    async def serve_forever(self):
        """
        Runs the server indefinitely.
        """
        if not self._server:
            raise RuntimeError("Server is not running. Use 'async with Server(...)'.")

        async with self._server:
            await self._server.serve_forever()



from typing import Any, Mapping
from pyspec._connection.associative_array import (
    AssociativeArray,
    AssociativeArrayElement,
    get_associative_array_key,
    pack_associative_array_element,
    unpack_associative_array_element,
)
import pytest


def test_key_parsing():
    assert get_associative_array_key("../x") is None
    assert get_associative_array_key("../x[1]") == ("1", "")
    assert get_associative_array_key("../x[1][2]") == ("1", "2")
    assert get_associative_array_key("../x[hello][goodbye]") == ("hello", "goodbye")


def test_unpack():
    arr = AssociativeArray()
    arr["1"] = "one"
    assert unpack_associative_array_element("../x[1]", arr) == "one"
    assert unpack_associative_array_element("../x", arr) == arr


def compare_array(arr1: AssociativeArray, arr2: AssociativeArray) -> bool:
    if set(arr1.data.keys()) != set(arr2.data.keys()):
        return False
    for key in arr1.data.keys():
        if arr1.data[key] != arr2.data[key]:
            return False
    return True


def test_pack():
    a1 = AssociativeArray()
    a1["1"] = "one"
    assert compare_array(
        pack_associative_array_element("../x[1]", "one"),  # type: ignore
        a1,
    )
    a2 = AssociativeArray()
    a2[1, 2] = "one two"
    assert compare_array(
        pack_associative_array_element("../x[1][2]", "one two"),  # type: ignore
        a2,
    )
    assert pack_associative_array_element("../x", "value") == "value"


def test_deleted():
    arr = AssociativeArray()
    arr["1"] = "one"
    arr["2"] = "two"
    arr["3"] = "three"
    del arr["2"]

    with pytest.raises(KeyError):
        arr["2"]

    arr[1]
    arr[3]





import sys
from enum import Enum
from typing import Literal, Union, Dict, Tuple
from .associative_array import AssociativeArray

import numpy as np

NATIVE_ENDIANNESS = "<" if sys.byteorder == "little" else ">"


class ErrorStr(str): ...


DataType = Union[np.ndarray, str, float, AssociativeArray, ErrorStr, None]


class Type(Enum):
    NONE = 0
    DOUBLE = 1
    STRING = 2
    ERROR = 3
    ASSOC = 4
    ARR_DOUBLE = 5
    ARR_FLOAT = 6
    ARR_LONG = 7
    ARR_ULONG = 8
    ARR_SHORT = 9
    ARR_USHORT = 10
    ARR_CHAR = 11
    ARR_UCHAR = 12
    ARR_STRING = 13
    ARR_LONG64 = 14
    ARR_ULONG64 = 15

    def is_array_type(self) -> bool:
        """
        Returns True if the type is an array type.

        Returns:
            bool: True if array type, False otherwise.
        """
        return self in {
            Type.ARR_DOUBLE,
            Type.ARR_FLOAT,
            Type.ARR_LONG,
            Type.ARR_ULONG,
            Type.ARR_SHORT,
            Type.ARR_USHORT,
            Type.ARR_CHAR,
            Type.ARR_UCHAR,
            Type.ARR_LONG64,
            Type.ARR_ULONG64,
            Type.ARR_STRING,
        }

    def to_numpy_type(
        self, endianness: Literal["<", ">"] = NATIVE_ENDIANNESS
    ) -> np.dtype:
        """
        Returns the numpy dtype corresponding to this SPEC type.

        Args:
            endianness (Literal['<', '>']): Endianness, '<' for little-endian, '>' for big-endian.
        Returns:
            np.dtype: Corresponding numpy dtype.
        Raises:
            ValueError: If the type is not an array type.
        """
        if self in ARRAY_TYPE_TO_NUMERIC_DTYPE:
            return with_endianness(
                ARRAY_TYPE_TO_NUMERIC_DTYPE[self], endianness=endianness
            )
        else:
            raise ValueError(f"Type {self} is not an array type.")

    @staticmethod
    def from_numpy_type(dtype: np.dtype) -> "Type":
        """
        Returns the SPEC Type corresponding to a numpy dtype.

        Args:
            dtype (np.dtype): Numpy dtype to convert.
        Returns:
            Type: Corresponding SPEC Type.
        Raises:
            ValueError: If the dtype is not supported.
        """
        if dtype.kind == "U":  # String type
            return Type.ARR_STRING
        if dtype.kind == "S":
            raise ValueError(
                "Byte string arrays are not supported. Ensure your strings are UTF-8 encoded."
            )
        key = (dtype_str(dtype), dtype.itemsize)
        if key in NUMERIC_DTYPE_TO_TYPE:
            return NUMERIC_DTYPE_TO_TYPE[key]
        raise ValueError(f"Unsupported numpy dtype: {dtype}")


DtypeStr = Literal["float", "int", "uint"]


def with_endianness(dtype: np.dtype, endianness: Literal["<", ">"]) -> np.dtype:
    """
    Returns a numpy dtype with the specified endianness.

    Args:
        dtype (np.dtype): The base numpy dtype.
        endianness (Literal['<', '>']): Endianness, '<' for little-endian, '>' for big-endian.
    Returns:
        np.dtype: Numpy dtype with specified endianness.
    """
    return np.dtype(endianness + dtype.str[1:])


def is_signed_int(dtype: np.dtype) -> bool:
    """
    Returns True if the dtype is a signed integer type.

    Args:
        dtype (np.dtype): Numpy dtype to check.
    Returns:
        bool: True if signed integer, False otherwise.
    """
    return np.issubdtype(dtype, np.signedinteger)


def is_floating_point(dtype: np.dtype) -> bool:
    """
    Returns True if the dtype is a floating point type.

    Args:
        dtype (np.dtype): Numpy dtype to check.
    Returns:
        bool: True if floating point, False otherwise.
    """
    return np.issubdtype(dtype, np.floating)


def dtype_str(dtype: np.dtype) -> DtypeStr:
    """
    Returns a string representing the type of the numpy dtype ('float', 'int', or 'uint').

    Args:
        dtype (np.dtype): Numpy dtype to check.
    Returns:
        DtypeStr: String representing the dtype category.
    Raises:
        ValueError: If the dtype is not supported.
    """
    if is_floating_point(dtype):
        return "float"
    elif is_signed_int(dtype):
        return "int"
    elif np.issubdtype(dtype, np.unsignedinteger):
        return "uint"
    else:
        raise ValueError(f"Unsupported dtype: {dtype}")


NUMERIC_DTYPE_TO_TYPE: Dict[Tuple[DtypeStr, int], Type] = {
    ("int", 1): Type.ARR_CHAR,
    ("uint", 1): Type.ARR_UCHAR,
    ("int", 2): Type.ARR_SHORT,
    ("uint", 2): Type.ARR_USHORT,
    ("int", 4): Type.ARR_LONG,
    ("uint", 4): Type.ARR_ULONG,
    ("float", 4): Type.ARR_FLOAT,
    ("int", 8): Type.ARR_LONG64,
    ("uint", 8): Type.ARR_ULONG64,
    ("float", 8): Type.ARR_DOUBLE,
}
ARRAY_TYPE_TO_NUMERIC_DTYPE: Dict[Type, np.dtype] = {
    Type.ARR_FLOAT: np.dtype(np.float32),
    Type.ARR_DOUBLE: np.dtype(np.float64),
    Type.ARR_CHAR: np.dtype(np.int8),
    Type.ARR_UCHAR: np.dtype(np.uint8),
    Type.ARR_SHORT: np.dtype(np.int16),
    Type.ARR_USHORT: np.dtype(np.uint16),
    Type.ARR_LONG: np.dtype(np.int32),
    Type.ARR_ULONG: np.dtype(np.uint32),
    Type.ARR_LONG64: np.dtype(np.int64),
    Type.ARR_ULONG64: np.dtype(np.uint64),
}



import ctypes
import pytest
import struct
import numpy as np
from pyspec._connection.protocol import (
    Header,
    Command,
    Type,
    serialize,
    _read_one_message,
    NATIVE_ENDIANNESS,
)
from pyspec._connection.data import AssociativeArray, ErrorStr
import asyncio
import itertools


async def read_one_message(
    msg: bytes,
):
    stream = asyncio.StreamReader()
    stream.feed_data(msg)
    stream.feed_eof()
    return await _read_one_message(stream)


@pytest.mark.asyncio
async def test_string_serialization():
    header = Header(Command.CMD, 0, "here is a name or something!")
    data = "some data"

    header_struct, data_bytes = serialize(header, data)

    read_header, read_data, endianness = await read_one_message(
        bytes(header_struct) + data_bytes
    )

    assert read_header.command == header.command
    assert read_header.name == header.name
    assert read_data == data
    assert endianness == NATIVE_ENDIANNESS


@pytest.mark.asyncio
async def test_associative_array_serialization():
    aa = AssociativeArray()
    aa["one"] = "now"
    aa["three"] = "the"
    aa["three", "0"] = "time"
    aa["two"] = "is"
    aa["four"] = "fourth"
    aa["four", "sub"] = "fourth"

    header = Header(Command.CMD, 0, "assoc array test")
    header_struct, data_bytes = serialize(header, aa)

    read_header, read_data, _ = await read_one_message(
        bytes(header_struct) + data_bytes
    )

    assert read_header.command == header.command
    assert read_header.name == header.name
    assert isinstance(read_data, AssociativeArray)
    assert read_data["one"] == "now"
    assert read_data["two"] == "is"
    assert read_data["three"] == "the"
    assert read_data["three", "0"] == "time"


Endianness = [
    "<",
    ">",
]

DTypes = [
    np.uint8,
    np.int8,
    np.uint16,
    np.int16,
    np.uint32,
    np.int32,
    np.float32,
    np.float64,
]


@pytest.mark.parametrize(
    "endianness, dtype",
    itertools.product(Endianness, DTypes),
)
@pytest.mark.asyncio
async def test_numpy_array_serialization(endianness, dtype):
    array = np.array([[1, 2, 3], [4, 5, 6]], dtype=dtype)
    header = Header(Command.CHAN_SEND, 1, f"numpy array test {dtype}")
    header_struct, data_bytes = serialize(header, array, endianness=endianness)

    assert header_struct.data_type == Type.from_numpy_type(array.dtype).value
    read_header, read_data, read_endianness = await read_one_message(
        bytes(header_struct) + data_bytes
    )

    assert read_header == header
    assert read_endianness == endianness
    assert isinstance(read_data, np.ndarray)
    np.testing.assert_array_equal(read_data, array)


@pytest.mark.asyncio
async def test_error_string_serialization():
    error_message = ErrorStr("This is an error.")

    header = Header(Command.CMD, 0, "error string test")
    header_struct, data_bytes = serialize(header, error_message)

    read_header, read_data, _ = await read_one_message(
        bytes(header_struct) + data_bytes
    )

    assert read_header == header
    assert isinstance(read_data, ErrorStr)
    assert read_data == error_message


@pytest.mark.asyncio
async def test_empty_data_serialization():
    header = Header(Command.CMD, 0, "empty data test")
    data = None

    header_struct, data_bytes = serialize(header, data)

    read_header, read_data, _ = await read_one_message(
        bytes(header_struct) + data_bytes
    )

    assert read_header == header
    # For empty data, we expect an empty string
    assert read_data == ""


@pytest.mark.asyncio
async def test_float_data_serialization():
    float_data = 3.14159

    header = Header(Command.CMD, 0, "float data test")
    header_struct, data_bytes = serialize(header, float_data)

    read_header, read_data, _ = await read_one_message(
        bytes(header_struct) + data_bytes
    )

    assert read_header == header
    assert isinstance(read_data, float)
    assert read_data == float_data


@pytest.mark.asyncio
async def test_int_data_serialization():
    int_data = 42

    header = Header(Command.REPLY, 0, "int data test")
    header_struct, data_bytes = serialize(header, int_data)

    read_header, read_data, _ = await read_one_message(
        bytes(header_struct) + data_bytes
    )

    assert read_header == header
    assert isinstance(read_data, int)
    assert read_data == int_data


@pytest.mark.asyncio
async def test_numpy_string_array_serialization():
    array = np.array([["foo", "bar"], ["baz", "qux"]], dtype="U10")
    header = Header(Command.CHAN_SEND, 2, "numpy string array test")
    header_struct, data_bytes = serialize(header, array)

    read_header, read_data, _ = await read_one_message(
        bytes(header_struct) + data_bytes
    )

    assert read_header == header
    assert isinstance(read_data, np.ndarray)
    assert read_data.dtype.kind in {"U", "S"}  # Unicode or bytes string
    np.testing.assert_array_equal(read_data, array)



from enum import Enum


class Command(Enum):
    """
    Enumeration of commands that can be sent between the client and server.

    This enum defines all possible commands for the SPEC protocol.
    """

    ### From Client ###
    CLOSE = 1
    ABORT = 2
    CMD = 3
    CMD_WITH_RETURN = 4

    REGISTER = 6
    UNREGISTER = 7

    FUNC = 9
    FUNC_WITH_RETURN = 10
    CHAN_READ = 11
    CHAN_SEND = 12

    HELLO = 14

    ### From Server ###
    EVENT = 8
    REPLY = 13
    HELLO_REPLY = 15

    ### Unused ###
    RETURN = 5



import ctypes
import pytest
import asyncio
from pyspec._connection.protocol import (
    HeaderV2_LE,
    HeaderV3_LE,
    HeaderV4_LE,
    SPEC_MAGIC,
    Command,
    Type,
    NAME_LEN,
    _read_one_message,
)


async def parse_header_and_data(header_bytes, data_bytes):
    stream = asyncio.StreamReader()
    stream.feed_data(header_bytes)
    stream.feed_data(data_bytes)
    stream.feed_eof()
    return await _read_one_message(stream)


@pytest.mark.asyncio
async def test_parse_old_headers():
    data2 = "abcd"
    data_bytes2 = data2.encode("utf-8") + b"\x00"
    h2 = HeaderV2_LE(
        magic=SPEC_MAGIC,
        version=2,
        size=ctypes.sizeof(HeaderV2_LE),
        length=len(data_bytes2),
        name="v2old".encode("utf-8").ljust(NAME_LEN, b"\x00"),
        data_type=Type.STRING.value,
        command=Command.CMD.value,
    )
    header2, data2_out, _ = await parse_header_and_data(bytes(h2), data_bytes2)
    assert header2.name == "v2old"
    assert data2_out == data2

    data3 = "12345j"
    data_bytes3 = data3.encode("utf-8") + b"\x00"
    h3 = HeaderV3_LE(
        magic=SPEC_MAGIC,
        version=3,
        size=ctypes.sizeof(HeaderV3_LE),
        length=len(data_bytes3),
        name="v3old".encode("utf-8").ljust(NAME_LEN, b"\x00"),
        data_type=Type.STRING.value,
        command=Command.CMD.value,
    )
    header3, data3_out, _ = await parse_header_and_data(bytes(h3), data_bytes3)
    assert header3.name == "v3old"
    assert data3_out == data3


@pytest.mark.asyncio
async def test_parse_newer_header_with_extra_fields():

    # V6 = V4 + extra fields (simulate by padding)
    class HeaderV6_LE(ctypes.LittleEndianStructure):
        _fields_ = tuple(HeaderV4_LE._fields_) + (
            ("extra1", ctypes.c_uint),
            ("extra2", ctypes.c_uint),
        )

    data6 = "abcdef"
    data_bytes = data6.encode("utf-8") + b"\x00"
    h6 = HeaderV6_LE(
        magic=SPEC_MAGIC,
        version=6,
        size=ctypes.sizeof(HeaderV6_LE),
        length=len(data_bytes),
        name="v6new".encode("utf-8").ljust(NAME_LEN, b"\x00"),
        data_type=Type.STRING.value,
        command=Command.CMD.value,
    )
    header_bytes6 = bytes(h6)
    header6, data6_out, _ = await parse_header_and_data(header_bytes6, data_bytes)
    assert header6.name == "v6new"
    assert data6_out == data6



# ******************************************************************************
#
#  @(#)css_logger.py	6.3  12/13/20 CSS
#
#  "pyspec" Release 6
#
#  Copyright (c) 2013,2014,2015,2016,2017,2018,2020
#  by Certified Scientific Software.
#  All rights reserved.
#
#  Permission is hereby granted, free of charge, to any person obtaining a
#  copy of this software ("pyspec") and associated documentation files (the
#  "Software"), to deal in the Software without restriction, including
#  without limitation the rights to use, copy, modify, merge, publish,
#  distribute, sublicense, and/or sell copies of the Software, and to
#  permit persons to whom the Software is furnished to do so, subject to
#  the following conditions:
#
#  The above copyright notice and this permission notice shall be included
#  in all copies or substantial portions of the Software.
#
#  Neither the name of the copyright holder nor the names of its contributors
#  may be used to endorse or promote products derived from this software
#  without specific prior written permission.
#
#     * The software is provided "as is", without warranty of any   *
#     * kind, express or implied, including but not limited to the  *
#     * warranties of merchantability, fitness for a particular     *
#     * purpose and noninfringement.  In no event shall the authors *
#     * or copyright holders be liable for any claim, damages or    *
#     * other liability, whether in an action of contract, tort     *
#     * or otherwise, arising from, out of or in connection with    *
#     * the software or the use of other dealings in the software.  *
#
# ******************************************************************************

import logging
from logging.handlers import RotatingFileHandler

import os
import time


class CSSLogger(logging.Logger):
    def isEnabledFor(self, level):
        return level <= self.getEffectiveLevel()

    def start(self, level=2, filename=None):
        if filename is not None:
            addFileHandler(filename)
        else:
            addStdOutHandler()
        self.setLevel(level)


class StdOutFormatter(logging.Formatter):

    def format(self, record):
        millis = int(round(record.created % 1 * 1000))
        strtime = time.strftime("%H:%M:%S", time.localtime(record.created))
        strtime += ".%03d" % millis
        basefile = os.path.basename(record.pathname)
        level = record.levelname
        levelno = record.levelno
        lineno = record.lineno
        funcname = record.funcName
        msg = record.msg

        fileinfo = "%s:%s (line:%s)" % (basefile, funcname, lineno)
        logline = "%s - %s - %-50s  | %s" % (strtime, levelno, fileinfo, msg)

        return logline


def start_log(level=2, filename=None):
    log.start(level, filename)


def addStdOutHandler():
    stdh = logging.StreamHandler()
    log_formatter = StdOutFormatter()
    stdh.setFormatter(log_formatter)
    log.addHandler(stdh)


def addFileHandler(filename):
    fileh = RotatingFileHandler(filename, maxBytes=2000000, backupCount=5)
    log_formatter = StdOutFormatter()
    fileh.setFormatter(log_formatter)
    log.addHandler(fileh)


def log_exception():
    import traceback

    log = CSSLogger("pyspec")
    log.log(2, traceback.format_exc())


log = CSSLogger("pyspec")



"""
This module implements the SPEC protocol for communication between
a client and server. It includes functionality for serializing and
deserializing messages, handling different protocol versions, and
managing endianness.

Most of the complexity arises from implementing compatibility with clients of
arbitrary endianness.

References:
- SPEC Server Help: https://certif.com/spec_help/server.html
- Associative Arrays: https://certif.com/spec_manual/ref_2_3_4_1.html

Notes:
The SPEC server protocol's has permissive endianness handling:
    "The spec server will check the endianess of the magic element of the first packet sent by the
    client and swap header and data bytes in that packet and subsequent incoming and outgoing data,
    if necessary, to accommodate the client. The client can send and read packet headers (and binary array data)
    using the native endian format of the client's platform."

So we have to be able to handle both big-endian and little-endian messages.
Endianness only affects multi-byte data types, which includes all numeric types.
Strings and string arrays are unaffected, since they are encoded in UTF-8, which is a single
byte encoding.



Classes:
- Header: Represents a message header.

Methods:
- serialize: Serializes a Header and DataType into bytes for sending.
- message_stream: Converts a StreamReader into an async generator of (Header, DataType) tuples.
- short_str: Constructs a short string representation of the header and data.
- long_str: Constructs a long string representation of the header and data.
"""

import asyncio
import ctypes
import logging
import struct
import time
from dataclasses import dataclass
from typing import Literal, TypeVar, Union
from .flag import Flag

import numpy as np

from .command import Command
from .associative_array import AssociativeArray, try_cast
from .data import (
    NATIVE_ENDIANNESS,
    AssociativeArray,
    DataType,
    ErrorStr,
    Type,
    with_endianness,
)

LOGGER = logging.getLogger("pyspec.protocol")

NAME_LEN = 80
SPEC_MAGIC = 4277009102


def log_once(logger: logging.Logger, level: int, msg: str):
    """
    Logs a message only once.

    Args:
        logger (logging.Logger): Logger instance to use.
        level (int): Logging level.
        msg (str): Message to log.
    """
    if not hasattr(log_once, "_logged_messages"):
        setattr(log_once, "_logged_messages", set())

    logged: set = getattr(log_once, "_logged_messages")
    if msg not in logged:
        logger.log(level, msg)
        logged.add(msg)


@dataclass
class Header:
    command: Command
    sequence_number: int = 0
    name: str = ""


def _encode_numpy_string_array(arr: np.ndarray) -> bytes:
    """
    Encode a numpy array of strings into bytes, with each string NULL-terminated.
    Numpy stores them as fixed-length, but SPEC expects NULL-terminated strings.

    Args:
        arr (np.ndarray): Numpy array of strings to encode.
    Returns:
        bytes: Encoded bytes with NULL-terminated strings.
    """
    byte_strings = [str(s).encode("utf-8") + b"\x00" for s in arr.flat]
    return b"".join(byte_strings)


def _decode_to_numpy_string_array(data_bytes: bytes) -> np.ndarray:
    """
    Decode bytes into a numpy array of strings, splitting on NULL bytes.
    Needs special handling since strings are not fixed-length.

    Args:
        data_bytes (bytes): Bytes to decode.
    Returns:
        np.ndarray: Numpy array of decoded strings with dtype '|U{max_length}', where max_length is the length of the longest string.
    """
    strings = [string.decode("utf-8") for string in data_bytes.split(b"\x00") if string]
    max_length = max(map(len, strings))
    array = np.array(strings, dtype=f"|U{max_length}")
    return array


FIELDS_V2 = [
    ("magic", ctypes.c_uint),
    ("version", ctypes.c_int),
    ("size", ctypes.c_int),
    ("sequence_number", ctypes.c_int),
    ("sec", ctypes.c_uint),
    ("usec", ctypes.c_uint),
    ("command", ctypes.c_int),
    ("data_type", ctypes.c_int),
    ("rows", ctypes.c_uint),
    ("cols", ctypes.c_uint),
    ("length", ctypes.c_uint),
    ("name", ctypes.c_char * NAME_LEN),
]


class HeaderV2_BE(ctypes.BigEndianStructure):
    _fields_ = FIELDS_V2


class HeaderV2_LE(ctypes.LittleEndianStructure):
    _fields_ = FIELDS_V2


FIELDS_V3 = [
    ("magic", ctypes.c_uint),
    ("version", ctypes.c_int),
    ("size", ctypes.c_int),
    ("sequence_number", ctypes.c_int),
    ("sec", ctypes.c_uint),
    ("usec", ctypes.c_uint),
    ("command", ctypes.c_int),
    ("data_type", ctypes.c_int),
    ("rows", ctypes.c_uint),
    ("cols", ctypes.c_uint),
    ("length", ctypes.c_uint),
    ("err", ctypes.c_int),
    ("name", ctypes.c_char * NAME_LEN),
]


class HeaderV3_BE(ctypes.BigEndianStructure):
    _fields_ = FIELDS_V3


class HeaderV3_LE(ctypes.LittleEndianStructure):
    _fields_ = FIELDS_V3


FIELDS_V4 = [
    ("magic", ctypes.c_uint),
    ("version", ctypes.c_int),
    ("size", ctypes.c_uint),  # Warning. Changed to unsigned in V4.
    ("sequence_number", ctypes.c_uint),
    ("sec", ctypes.c_uint),
    ("usec", ctypes.c_uint),
    ("command", ctypes.c_int),
    ("data_type", ctypes.c_int),
    ("rows", ctypes.c_uint),
    ("cols", ctypes.c_uint),
    ("length", ctypes.c_uint),
    ("err", ctypes.c_int),
    ("flags", ctypes.c_int),
    ("name", ctypes.c_char * NAME_LEN),
]


class HeaderV4_BE(ctypes.BigEndianStructure):
    _fields_ = FIELDS_V4


class HeaderV4_LE(ctypes.LittleEndianStructure):
    _fields_ = FIELDS_V4


HeaderStruct = Union[
    HeaderV2_BE,
    HeaderV2_LE,
    HeaderV3_BE,
    HeaderV3_LE,
    HeaderV4_BE,
    HeaderV4_LE,
]

S = TypeVar("S", bound=ctypes.Structure)


HEADER_STRUCT_MAP = {
    (2, "<"): HeaderV2_LE,
    (2, ">"): HeaderV2_BE,
    (3, "<"): HeaderV3_LE,
    (3, ">"): HeaderV3_BE,
    (4, "<"): HeaderV4_LE,
    (4, ">"): HeaderV4_BE,
}
LE_HEADERS = (HeaderV4_LE, HeaderV3_LE, HeaderV2_LE)
BE_HEADERS = (HeaderV4_BE, HeaderV3_BE, HeaderV2_BE)


def _determine_header_struct(
    version: int,
    header_size: int,
    endianness: Literal["<", ">"] = "<",
) -> type[HeaderStruct]:
    """
    Attempts to determine the appropriate Header structure based on version number, header size, and apparent endianness.

    Args:
        version (int): Header version number.
        header_size (int): Size of the header in bytes.
        endianness (Literal['<', '>']): Endianness, '<' for little-endian, '>' for big-endian.
    Returns:
        type[HeaderStruct]: Header structure class.
    Raises:
        RuntimeError: If a suitable header structure cannot be determined.
    """

    header_struct = HEADER_STRUCT_MAP.get((version, endianness), None)
    if header_struct is not None:
        return header_struct

    # If we don't have a known version, then just try to pick the target based on size.
    # We can't safely deserialize a header if it requires more bytes than we have.
    header_options = LE_HEADERS if endianness == "<" else BE_HEADERS
    for header_struct in header_options:
        if ctypes.sizeof(header_struct) <= header_size:
            log_once(
                LOGGER,
                logging.WARNING,
                f"Unknown header version {version}; "
                f"falling back to header struct {header_struct.__name__} based on size {header_size}.",
            )
            return header_struct

    if header_struct is None:
        raise RuntimeError(
            f"Could not safely deserialize header with size {header_size} and version {version}; size too small."
        )
    return header_struct


def _deserialize_data(
    header: HeaderStruct,
    data_bytes: bytes,
    endianness: Literal["<", ">"] = NATIVE_ENDIANNESS,
) -> DataType:
    """
    Deserialize the given bytes into data.
    Uses the type, rows, and cols attributes to determine how to deserialize.

    Args:
        header (HeaderStruct): Header structure containing type and shape information.
        data_bytes (bytes): Bytes to deserialize.
        endianness (Literal['<', '>']): Endianness, '<' for little-endian, '>' for big-endian.
    Returns:
        DataType: Deserialized data.
    Raises:
        ValueError: If the type is unsupported for deserialization.
        UnicodeDecodeError: If the bytes cannot be decoded as UTF-8 for string types.
    """
    if data_bytes:
        assert data_bytes.endswith(b"\x00"), "Data bytes should end with NULL byte."
        data_bytes = data_bytes[:-1]

    data_type = Type(header.data_type)
    if data_type == Type.DOUBLE:
        # This is not actually sent by a true SPEC server.
        return struct.unpack(f"{endianness}d", data_bytes)[0]
    elif data_type == Type.STRING:
        data_string = data_bytes.decode("utf-8")
        # Try to cast to numeric is possible.
        # https://certif.com/spec_help/server.html
        # The spec server sends both string-valued and number-valued items as strings.
        # Numbers are converted to strings using a printf("%.15g") format.
        # However, spec will accept Type.DOUBLE values.
        value = try_cast(data_string)
        return value
    elif data_type == Type.ASSOC:
        deleted = header.version >= 4 and (header.flags & Flag.DELETED.value) != 0
        return AssociativeArray.deserialize(data_bytes, deleted)
    elif data_type == Type.ERROR:
        return ErrorStr(data_bytes.decode("utf-8"))
    elif data_type.is_array_type():
        if data_type == Type.ARR_STRING:
            array = _decode_to_numpy_string_array(data_bytes)
        else:
            array = np.frombuffer(data_bytes, dtype=data_type.to_numpy_type(endianness))

        array = array.reshape((header.rows, header.cols))
        # TODO: Need to test this with SPEC.
        # I am not sure whether SPEC would send a 0 or 1 for the other dimension of a 1D array.
        # Based on the old pyspec impl, it looks like vectors are always row vectors.
        # So if we have rows == 1, we flatten to 1D.
        if header.rows == 1:
            array = array.flatten()
        return array


async def _read_prefix(
    stream: asyncio.StreamReader,
):
    """
    Reads the header prefix from the stream.

    Args:
        stream (asyncio.StreamReader): The stream to read from.
    Returns:
        tuple: The raw prefix bytes, version, size, and endianness.
    Raises:
        RuntimeError: If the magic number is invalid.
    """

    # Warning. We parse the size field as an unsigned int here, since in V4 it was changed to unsigned.
    # However, older versions of SPEC used a signed int.
    # Unless the size of the header is larger than 2GB, this should not cause issues.
    fields = "Iii"
    prefix_bytes = await stream.readexactly(struct.calcsize(fields))
    (magic_le, version, size) = struct.unpack(f"<{fields}", prefix_bytes)
    if magic_le == SPEC_MAGIC:
        endianness = "<"
        return prefix_bytes, version, size, endianness

    (magic_be, version, size) = struct.unpack(f">{fields}", prefix_bytes)
    if magic_be == SPEC_MAGIC:
        endianness = ">"
        return prefix_bytes, version, size, endianness

    raise RuntimeError(
        f"Invalid magic number: {magic_le} (LE) / {magic_be} (BE); expected {SPEC_MAGIC}."
    )


async def _read_one_message(
    stream: asyncio.StreamReader,
    logger: logging.Logger = LOGGER,
) -> tuple[Header, DataType, Literal["<", ">"]]:
    """
    Attempts to read a single message (header + data) from the stream.

    Args:
        stream (asyncio.StreamReader): The stream to read from.
        logger (logging.Logger): Logger for logging messages.
    Returns:
        tuple: The header, deserialized data, and endianness.
    Raises:
        RuntimeError: If the magic number is invalid.
    """
    prefix_bytes, version, header_size, apparent_endianness = await _read_prefix(stream)

    header_struct = _determine_header_struct(version, header_size, apparent_endianness)
    header = header_struct.from_buffer_copy(
        prefix_bytes
        + await stream.readexactly(
            max(ctypes.sizeof(header_struct), header_size) - len(prefix_bytes)
        )
    )

    data_bytes = await stream.readexactly(header.length)
    data = _deserialize_data(header, data_bytes, apparent_endianness)

    logger.info("Received: %s", short_str(header, data))
    logger.debug("Detail: %s", long_str(header, data))
    logger.debug("Raw data bytes: %s", data_bytes)

    return (
        Header(
            command=Command(header.command),
            sequence_number=header.sequence_number,
            name=header.name.decode("utf-8").rstrip("\x00"),
        ),
        data,
        apparent_endianness,
    )


async def message_stream(
    stream: asyncio.StreamReader,
    logger: logging.Logger = LOGGER,
):
    """
    Converts a StreamReader into an async generator of (Header, DataType) tuples.

    Args:
        stream (asyncio.StreamReader): The stream to read from.
        logger (logging.Logger): Logger for logging messages.
    Yields:
        tuple: The header, deserialized data, and endianness.
    """
    while True:
        try:
            yield await _read_one_message(stream, logger)
        except asyncio.IncompleteReadError:
            break


def serialize(
    header: Header,
    data: DataType,
    endianness: Union[Literal["<", ">"], None] = NATIVE_ENDIANNESS,
) -> tuple[HeaderStruct, bytes]:
    """
    Serializes a Header and DataType into bytes for sending.

    Args:
        header (Header): The header to serialize.
        data (DataType): The data to serialize.
        endianness (Literal['<', '>']): Endianness, '<' for little-endian, '>' for big-endian, or None for native.
    Returns:
        tuple: The serialized header structure and data bytes.
    Raises:
        ValueError: If the data type is not supported for serialization.
    """
    if endianness is None:
        endianness = NATIVE_ENDIANNESS

    rows, cols = 0, 0
    data_bytes = b""
    if isinstance(data, ErrorStr):
        data_type = Type.ERROR
        data_bytes = data.encode("utf-8")
    elif isinstance(data, (str, int, float)):
        data_type = Type.STRING
        # The spec server sends both string-valued and number-valued items as strings.
        # Numbers are converted to strings using a printf("%.15g") format.

        # Despite SPEC claiming that it will handle DOUBLE type things just fine,
        # we have found that its better to just keep them STRINGs, since that is what it really
        # expects. Especially if it is going to concatenate it to a string in a command, which is common.
        if isinstance(data, (int, float)):
            # Note: This happens to convert bools to 1/0 strings.
            # SPEC wants them in 1/0 strings anyway.
            data = f"{data:.15g}"
        data_bytes = struct.pack("{}s".format(len(data)), data.encode("utf-8"))
    elif isinstance(data, AssociativeArray):
        data_type = Type.ASSOC
        data_bytes = data.serialize()
    elif isinstance(data, np.ndarray):
        data_type = Type.from_numpy_type(data.dtype)
        if data.ndim > 2:
            raise ValueError("Only 1D and 2D arrays are supported.")
        data = np.atleast_2d(data)

        rows, cols = data.shape
        if data_type == Type.ARR_STRING:
            data_bytes = _encode_numpy_string_array(data)
        else:
            # Make sure the data is encoded in the right format.
            # This is not relevant for string arrays.
            # String arrays are communicated as a sequence of 1 byte chars (utf-8).
            data_bytes = data.astype(
                with_endianness(data.dtype, endianness), copy=False
            ).tobytes()
    elif data is None:
        data_type = Type.STRING  # Default to STRING type for None
    else:
        raise ValueError(f"Cannot serialize data of type {type(data)}.")

    if data_bytes:
        # TODO: This is a silly copy.
        data_bytes += b"\x00"

    # Send as V4 header.
    header_struct = HeaderV4_LE if endianness == "<" else HeaderV4_BE

    return (
        header_struct(
            magic=SPEC_MAGIC,
            version=4,
            size=ctypes.sizeof(header_struct),
            sequence_number=header.sequence_number,
            sec=int(time.time()),
            usec=int((time.time() % 1) * 1_000_000),
            command=header.command.value,
            data_type=data_type.value,
            rows=rows,
            cols=cols,
            length=len(data_bytes),
            err=0,
            flags=0,
            name=header.name.encode("utf-8").ljust(NAME_LEN, b"\x00"),
        ),
        data_bytes,
    )


def short_str(header: HeaderStruct, data: DataType) -> str:
    """
    Construct a short string representation of the header and data.
    Focuses on the command and key parameters.
    Useful for frequent logging compared to ``long_str``.

    Args:
        header (HeaderStruct): The header to represent.
        data (DataType): The data to represent.
    Returns:
        str: Short string representation.
    """
    cmd = Command(header.command)
    name = header.name.decode("utf-8").rstrip("\x00")

    if cmd == Command.HELLO:
        return f"{cmd.name}(seq={header.sequence_number})"
    if cmd == Command.HELLO_REPLY:
        return f"{cmd.name}(seq={header.sequence_number})"
    if cmd in (Command.CMD, Command.FUNC):
        return f"{cmd.name}(`{data}`)"
    if cmd in (Command.CMD_WITH_RETURN, Command.FUNC_WITH_RETURN):
        return f"{cmd.name}(`{data}`, seq={header.sequence_number})"
    if cmd == Command.CHAN_SEND:
        return f"{cmd.name}(`{name}`, seq={header.sequence_number})"
    if cmd in (Command.REGISTER, Command.UNREGISTER):
        return f"{cmd.name}(`{name}`)"
    if cmd == Command.EVENT:
        return f"{cmd.name}(`{name}`)"
    if cmd == Command.REPLY:
        return f"{cmd.name}(seq={header.sequence_number})"
    if cmd in (Command.CLOSE, Command.ABORT):
        return f"{cmd.name}"
    if cmd == Command.RETURN:
        # This is unused in the current version of SPEC.
        # So the semantics are not defined.
        return f"{cmd.name}(seq={header.sequence_number})"
    else:
        return f"{cmd.name}(seq={header.sequence_number})"


def long_str(header: HeaderStruct, data: DataType) -> str:
    """
    Construct a long string representation of the header and data.
    Contains all of the information in the V2 header.

    Args:
        header (HeaderStruct): The header to represent.
        data (DataType): The data to represent.
    Returns:
        str: Long string representation.
    """
    cmd = Command(header.command)
    _type = Type(header.data_type)
    return (
        f"<Header version={header.version} magic={header.magic} size={header.size} cmd={cmd.name} "
        f"type={_type.name} name='{header.name.decode('utf-8').rstrip(chr(0))}' seq={header.sequence_number} "
        f"rows={header.rows} cols={header.cols} length={header.length} data={data}>"
    )



import asyncio
import ctypes
import logging
from dataclasses import dataclass
from typing import Literal, TypeVar, overload, Optional, Union

from pyee.asyncio import AsyncIOEventEmitter


from . import protocol
from .protocol import DataType, Header, message_stream

LOGGER = logging.getLogger("pyspec.connection")


T = TypeVar("T", bound=ctypes.Structure)


class Connection(AsyncIOEventEmitter):
    host: str
    """The host address of the connection."""
    port: int
    """The port number of the connection."""

    @dataclass
    class Message:
        """
        Represents a message received from the connection.

        Args:
            header (Header): The message header.
            data (DataType): The message data.
        """

        header: Header
        data: DataType

    @overload
    def __init__(
        self,
        host: str,
        port: int,
        /,
    ) -> None:
        """
        Initialize a connection to a remote host.

        Args:
            host (str): The hostname or IP address of the remote server.
            port (int): The port number of the remote server.
        """

    @overload
    def __init__(
        self,
        reader: asyncio.StreamReader,
        writer: asyncio.StreamWriter,
        /,
    ) -> None:
        """
        Initialize a connection using existing StreamReader and StreamWriter.
        This version should be used when the socket connection is already established
        and you just want to interface with the sockets using the protocols defined here.

        Args:
            reader (asyncio.StreamReader): The StreamReader for the connection.
            writer (asyncio.StreamWriter): The StreamWriter for the connection.
        """

    def __init__(
        self,
        host_or_reader: Union[str, asyncio.StreamReader],
        port_or_writer: Union[int, asyncio.StreamWriter],
        /,
    ) -> None:
        """
        Initialize a connection to a remote host or using an existing stream.

        Args:
            host_or_reader (str or asyncio.StreamReader): Hostname or StreamReader.
            port_or_writer (int or asyncio.StreamWriter): Port number or StreamWriter.
        Raises:
            TypeError: If argument types are invalid.
        """
        super().__init__()

        if isinstance(host_or_reader, asyncio.StreamReader):
            if not isinstance(port_or_writer, asyncio.StreamWriter):
                raise TypeError(
                    "If the first argument is a StreamReader, the second must be a StreamWriter"
                )
            self._reader = host_or_reader
            self._writer = port_or_writer
            self.host: str = self._writer.get_extra_info("peername")[0]
            self.port: int = self._writer.get_extra_info("peername")[1]
        else:
            if not isinstance(port_or_writer, int):
                raise TypeError(
                    "If the first argument is a host string, the second must be an integer port"
                )
            self.host: str = host_or_reader
            self.port: int = port_or_writer
            self._reader = None
            self._writer = None

        self._listener: Optional[asyncio.Task] = None
        self.logger = LOGGER.getChild(f"{self.host}:{self.port}")

        self._peer_endianness: Optional[Literal["<", ">"]] = None
        """
        Keeps track of the endianness of the remote connection.
        Endianness is determined by looking at the magic number in the header of incoming messages.
        If None, endianness has not yet been determined, and messages will be sent with native endianness.

        If set, all messages sent along this connection will use the determined endianness.
        A warning will be logged if the endianness is changed mid-connection.
        """

    @property
    def is_connected(self) -> bool:
        """
        Returns True if the connection is established and open.

        Returns:
            bool: True if the connection is open, False otherwise.
        """
        return self._writer is not None and not self._writer.is_closing()

    async def __aenter__(self) -> "Connection":
        """
        Enter the async context manager for the connection.

        Returns:
            Connection: The connection instance.
        """
        self._listener = asyncio.create_task(self._listen())
        return self

    async def __aexit__(self, exc_type, exc, tb) -> None:
        """
        Exit the async context manager for the connection.
        Cancels the listener task if it exists.
        """
        if self._listener:
            try:
                self._listener.cancel()
            except asyncio.CancelledError:
                await self._listener

    async def __send(self, msg: bytes) -> None:
        """
        Sends raw data to the connected server.

        Args:
            msg (bytes): The bytes to send.
        """
        assert self._writer is not None, "Connection is not established."
        self.logger.debug("Sending message: %s", msg)
        self._writer.write(msg)
        await self._writer.drain()

    async def _send(self, header: Header, data: DataType = None) -> None:
        """
        Sends a message to the connected server.

        Args:
            header (Header): The header to send.
            data (DataType, optional): The data to send.
        """
        header_struct, data_bytes = protocol.serialize(
            header, data, self._peer_endianness
        )
        self.logger.info("Sending: %s", protocol.short_str(header_struct, data))
        self.logger.debug("Detail: %s", protocol.long_str(header_struct, data))

        await self.__send(bytes(header_struct))
        if data_bytes:
            await self.__send(data_bytes)

    async def _listen(self):
        """
        Listens for raw data from the connected server.
        Emits a 'message' event for each received message.
        """
        assert self._reader is not None, "Connection is not established."
        try:
            async for header, data, endianness in message_stream(
                self._reader, self.logger
            ):
                if self._peer_endianness is not None:
                    if self._peer_endianness != endianness:
                        self.logger.warning(
                            "Endianness changed mid-connection from %s to %s",
                            self._peer_endianness,
                            endianness,
                        )
                self._peer_endianness = endianness
                self.emit("message", Connection.Message(header, data))
        except asyncio.CancelledError:
            self.logger.info("Listener task cancelled, stopping listener.")
        except asyncio.IncompleteReadError:
            self.logger.info("Connection closed by peer.")
        except Exception as e:
            self.logger.exception("Error while listening for messages: %s", e)
            raise



from __future__ import annotations

import asyncio
import inspect
import logging
from collections import defaultdict
from contextlib import AsyncExitStack, asynccontextmanager
from pathlib import Path
from typing import (
    Any,
    AsyncIterator,
    Awaitable,
    Callable,
    Generic,
    Literal,
    TypeVar,
    cast,
)

from pyee.asyncio import AsyncIOEventEmitter
from typing_extensions import Self

from pyspec._connection import ClientConnection
from pyspec._connection.associative_array import (
    AssociativeArray,
    get_associative_array_key,
    pack_associative_array_element,
    unpack_associative_array_element,
)
from pyspec._connection.data import DataType

T = TypeVar("T", bound=DataType)
K = TypeVar("K", bound=DataType)
LOGGER = logging.getLogger(__name__)


class ContextWaiter:
    """
    This class allows for awaiting an awaitable either via

    .. code-block:: python

        async with ContextWaiter(...):
            ...

    or via

    .. code-block:: python

        await ContextWaiter(...)

    Args:
        awaitable (Awaitable): The awaitable to wait for.
    """

    def __init__(self, awaitable: Awaitable):
        self._awaitable = awaitable

    async def __aenter__(self) -> Self:
        return self

    async def __aexit__(self, exc_type, exc, tb) -> None:
        await self._awaitable

    def __await__(self):
        async def enter_exit():
            async with self:
                pass

        return enter_exit().__await__()


class RemotePropertyTable(AsyncIOEventEmitter):
    """
    Class to manage remote properties on a SPEC server.
    Allows reading, writing, and subscribing to property changes.

    Args:
        connection (ClientConnection): The client connection to the SPEC server.
    """

    def __init__(self, connection: ClientConnection):
        super().__init__()
        self._property_watchers: dict[str, int] = defaultdict(int)
        self._table: dict[str, DataType] = {}
        self._connection = connection
        self._connection.on("property-change", self._on_property_update)

    def _update_cache(self, property_name: str, value: DataType) -> None:
        # If we got an update from one associative array to another, we want to merge the updates.
        # We will only get incremental updates as they come.

        if (
            (existing := self._table.get(property_name, None)) is not None
            and isinstance(existing, AssociativeArray)
            and isinstance(value, AssociativeArray)
        ):
            existing.update(value)
        else:
            self._table[property_name] = value

    def _on_property_update(self, property_name: str, value: DataType) -> None:
        value = unpack_associative_array_element(property_name, value)
        self._update_cache(property_name, value)
        LOGGER.info("Dispatching property update for '%s'", property_name)
        self.emit(f"property-{property_name}", value)

    def _increment_watcher(self, property_name: str) -> bool:
        self._property_watchers[property_name] += 1
        return self._property_watchers[property_name] == 1

    def _decrement_watcher(self, property_name: str) -> bool:
        if property_name in self._property_watchers:
            self._property_watchers[property_name] -= 1
            if self._property_watchers[property_name] <= 0:
                del self._property_watchers[property_name]
                return True
        return False

    async def _read(self, property_name: str) -> DataType:
        value = await self._connection.prop_get(property_name)
        value = unpack_associative_array_element(property_name, value)
        return value

    async def read(self, property_name: str) -> DataType:
        """
        Read the value of a property from the local cache or from the server.
        If the property is not in the local cache, it will be fetched from the server.

        The result is only cached if it has been subscribed to.

        Args:
            property_name (str): The name of the property to read.
        Returns:
            DataType: The value of the property.
        """
        if self.is_subscribed(property_name):
            # This only happens when we are subscribed to the property
            # but there hasn't been a change since we subscribed.
            if property_name not in self._table:
                self._update_cache(property_name, await self._read(property_name))
            return self._table[property_name]

        return await self._read(property_name)

    async def read_next(self, property_name: str) -> DataType:
        """
        Wait for the next update of a property and return its value.
        See read() for more details.

        Args:
            property_name (str): The name of the property to read.
        Returns:
            DataType: The next value of the property.
        """
        assert self.is_subscribed(
            property_name
        ), "Property must be watched to read next value."

        future = asyncio.Future()
        self.once(f"property-{property_name}", lambda value: future.set_result(value))
        return await future

    async def write(self, property_name: str, value: DataType) -> None:
        """
        Writes a value to a property on the server.

        Args:
            property_name (str): The name of the property to write to.
            value (DataType): The value to write to the property.
        """
        self._update_cache(property_name, value)
        value = pack_associative_array_element(property_name, value)
        await self._connection.prop_set(property_name, value)

    async def subscribe(self, property_name: str) -> None:
        """
        Subscribe to updates for a property.

        Subscribing to a property will cause the property values to be
        cached locally and the provided callback to be called
        whenever the property value changes.

        Args:
            property_name (str): The name of the property to subscribe to.
        """
        if self._increment_watcher(property_name):
            await self._connection.prop_watch(property_name)

    async def unsubscribe(self, property_name: str) -> None:
        """
        Unsubscribes from a property.

        Args:
            property_name (str): The name of the property to unsubscribe from.
        """
        if self._decrement_watcher(property_name):
            await self._connection.prop_unwatch(property_name)

    def property(
        self,
        name: str,
        coerce: Callable[[Any], T] | None = None,
    ) -> Property[T]:
        """
        Gets a helper object to manage interfacing with a read-write property.

        Args:
            name (str): The property name.
            coerce (Callable[[Any], T], optional): Optional function to coerce the data to a specific type.
        Returns:
            Property[T]: A Property helper object.
        """
        return Property(name, self, coerce)

    def event_stream(
        self,
        name: str,
        coerce: Callable[[Any], T] | None = None,
    ) -> EventStream[T]:
        """
        Gets a helper object to manage interfacing with an event stream property.

        Args:
            name (str): The property name.
            coerce (Callable[[Any], T], optional): Optional function to coerce the data to a specific type.
        Returns:
            EventStream[T]: An EventStream helper object.
        """
        return EventStream(name, self, coerce)

    def readonly_property(
        self,
        name: str,
        coerce: Callable[[Any], T] | None = None,
    ) -> ReadableProperty[T]:
        """
        Gets a helper object to manage interfacing with a read-only property.

        Args:
            name (str): The property name.
            coerce (Callable[[Any], T], optional): Optional function to coerce the data to a specific type.
        Returns:
            ReadableProperty[T]: A ReadableProperty helper object.
        """
        return ReadableProperty(name, self, coerce)

    def writeonly_property(self, name: str) -> WritableProperty:
        """
        Gets a helper object to manage interfacing with a write-only property.

        Args:
            name (str): The property name.
        Returns:
            WritableProperty: A WritableProperty helper object.
        """
        return WritableProperty(name, self)

    def is_subscribed(self, property_name: str) -> bool:
        """
        Checks if there are any active subscriptions to a property.

        Args:
            property_name (str): The name of the property to check.
        Returns:
            bool: True if there are active subscriptions, False otherwise.
        """
        return (
            property_name in self._property_watchers
            and self._property_watchers[property_name] > 0
        )


class _PropertyBase(Generic[T]):
    def __init__(
        self,
        name: str,
        property_table: "RemotePropertyTable",
    ):
        super().__init__()
        self.name = name
        self._property_table = property_table


class EventStream(_PropertyBase[T]):
    """
    Represents a stream of events for a property. Events are sent from the server whenever the property value changes or when the server wants to send data to clients.

    Args:
        name (str): The property name.
        property_table (RemotePropertyTable): The remote property table instance.
        coerce (Callable[[DataType], T], optional): Optional function to coerce the data to a specific type.
    """

    EventType = Literal["change", "event"]

    _subscriber_depth = 0

    def __init__(
        self,
        name: str,
        property_table: "RemotePropertyTable",
        coerce: Callable[[DataType], T] | None = None,
    ):
        super().__init__(name, property_table)
        self._emitter = AsyncIOEventEmitter()
        self._coerce = coerce

    def _cast_value(self, value: DataType) -> T:
        if self._coerce is not None:
            try:
                return self._coerce(value)
            except Exception as e:
                raise TypeError(
                    f"Failed to coerce value '{value}' to type {self._coerce}"
                ) from e
        return cast(T, value)

    def _emit_change(self, value: T) -> None:
        try:
            value = self._cast_value(value)
            self._emitter.emit("change", value)
        except TypeError:
            LOGGER.error(
                "Received value of incorrect type for property '%s': expected %s, got %s",
                self.name,
                type(value),
            )
            return

    def _emit_event(self, value: T) -> None:
        self._emitter.emit("event", value)

    def on(self, event: EventStream.EventType, func: Callable[[T], None]) -> None:
        self._emitter.on(event, func)

    def once(self, event: EventStream.EventType, func: Callable[[T], None]) -> None:
        self._emitter.once(event, func)

    def remove_listener(
        self, event: EventStream.EventType, func: Callable[[T], None]
    ) -> None:
        self._emitter.remove_listener(event, func)

    async def get_next(self) -> T:
        """
        Waits for the next change event and then returns the value

        Returns:
            T: The value of the property after the next time it changes.
        """
        value = await self._property_table.read_next(self.name)
        return self._cast_value(value)

    def wait_for(self, value: T, *, timeout: float | None = None) -> ContextWaiter:
        """
        Waits until the property changes to the specified value.

        Usage:
        .. code-block:: python

            async with test_var.wait_for(123, timeout=10):
                # Do something...

            # Wait for the property to change to 123 before continuing

        .. code-block:: python

            # Just wait until the property changes
            await test_var.wait_for(123, timeout=10)

        Args:
            value (T): The value to wait for.
            timeout (float, optional): Optional timeout in seconds.
        """
        assert self.is_subscribed(), "Property must be watched to wait for a value."
        future = asyncio.Future()

        def check_value(new_value: T) -> None:
            if new_value == value and not future.done():
                future.set_result(None)

        def on_done(*args, **kwargs) -> None:
            self.remove_listener("change", check_value)

        future.add_done_callback(on_done)

        self.on("change", check_value)
        return ContextWaiter(asyncio.wait_for(future, timeout))

    def is_subscribed(self) -> bool:
        """
        Checks if there are any active subscriptions to this property.

        Returns:
            bool: True if there are active subscriptions, False otherwise.
        """
        return self._property_table.is_subscribed(self.name)

    @asynccontextmanager
    async def subscribed(
        self,
    ) -> AsyncIterator[Self]:
        """
        Context manager that subscribes to the property for the duration of the context.

        WARNING: Data arrays will not send events. You must use get() to receive the data array values.
        """

        event_name = f"property-{self.name}"
        initialized = False

        def skip_first_emit_change(value: T) -> None:
            # This shenanigans is necessary to preserve change semantics.
            # The initial value of the property is sent back immediately upon subscribing as an "event".
            # We don't want to call that a "change" at this level, so we don't emit that.
            # The initialized value is of course still available from read() at any time.
            nonlocal initialized
            if initialized:
                self._emit_change(value)
            initialized = True

        try:
            if self._subscriber_depth == 0:
                self._property_table.on(event_name, skip_first_emit_change)
                self._property_table.on(event_name, self._emit_event)
            self._subscriber_depth += 1
            await self._property_table.subscribe(self.name)
            yield self
        finally:
            self._subscriber_depth -= 1
            if self._subscriber_depth == 0:
                self._property_table.remove_listener(event_name, skip_first_emit_change)
                self._property_table.remove_listener(event_name, self._emit_event)
            await self._property_table.unsubscribe(self.name)

    @asynccontextmanager
    async def capture(self, wait_time: float = 0.0):
        """
        Context manager that captures output from the server for the duration of the context.

        Example usage:

        .. code-block:: python

            async with client.output("tty").capture() as output:
                await client.exec('print("Hello, world!")')
                assert output[-1] == "Hello, world!\\n"

        Args:
            wait_time (float, optional): Optional time to wait after the context block exits to ensure all output is captured.
        Yields:
            list: List of output lines captured during the context.
        """
        lines = []
        self.on("event", lines.append)
        try:
            async with self.subscribed():
                yield lines
                # Wait for any remaining output to be captured before unsubscribing.
                await asyncio.sleep(wait_time)
        finally:
            self.remove_listener("event", lines.append)


class ReadableProperty(EventStream[T]):
    async def get(self) -> T:
        value = await self._property_table.read(self.name)
        return self._cast_value(value)


class WritableProperty(_PropertyBase[T]):
    async def set(self, value: T) -> None:
        await self._property_table.write(self.name, value)


class Property(ReadableProperty[T], WritableProperty[T]): ...


class PropertyGroup:

    _stack = AsyncExitStack()

    def __init__(self, prefix: str | Path, remote_property_table: RemotePropertyTable):
        self._prefix = Path(prefix)
        self._remote_property_table = remote_property_table

    def _path(self, name: str) -> str:
        return (self._prefix / name).as_posix()

    def _property(
        self,
        name: str,
        coerce: Callable[[Any], T] | None = None,
    ) -> Property[T]:
        return self._remote_property_table.property(self._path(name), coerce)

    def _event_stream(
        self,
        name: str,
        coerce: Callable[[Any], T] | None = None,
    ) -> EventStream[T]:
        return self._remote_property_table.readonly_property(self._path(name), coerce)

    def _readonly_property(
        self,
        name: str,
        coerce: Callable[[Any], T] | None = None,
    ) -> ReadableProperty[T]:
        return self._remote_property_table.readonly_property(self._path(name), coerce)

    def _writeonly_property(
        self,
        name: str,
    ) -> WritableProperty:
        return self._remote_property_table.writeonly_property(self._path(name))

    async def __aenter__(self) -> Self:
        await self._stack.__aenter__()
        # Attempt to subscribe to all at the same time.
        await asyncio.gather(
            *[
                self._stack.enter_async_context(prop.subscribed())
                for _, prop in inspect.getmembers(self)
                # Only subscribe to ReadableProperty instances
                if isinstance(prop, ReadableProperty)
            ]
        )
        return self

    async def __aexit__(self, exc_type, exc, tb) -> None:
        await self._stack.__aexit__(exc_type, exc, tb)
        self._stack = AsyncExitStack()



from ._remote_property_table import PropertyGroup, RemotePropertyTable


class Status(PropertyGroup):
    """
    The status properties reflect changes in the server state that may affect the server's ability to execute client commands or control hardware.

    Args:
        remote_property_table (RemotePropertyTable): The remote property table instance.

    Attributes:
        quit (bool): Whether or not the server is exiting.
        shell (bool): Whether or not the server is in a subshell.
        ready (bool): Whether or not the server is ready to execute commands.
        simulate (bool): Whether or not the server is in simulate mode.
    """

    def __init__(self, remote_property_table: RemotePropertyTable):
        super().__init__("status", remote_property_table)
        self._remote_property_table = remote_property_table

        self.quit = self._readonly_property("quit", bool)
        """
        status/quit
        
        .. code-block:: none

            on("change"): Sent when the server exits.
            get: Always read back as zero.
        """

        self.shell = self._readonly_property("shell", bool)
        """
        status/shell

        .. code-block:: none

            on("change"): Sent when the server enters a subshell (True) or returns from a subshell (False).
            get: True if server command thread is in a subshell, otherwise False.

        Starting with spec release 6.06.01, when entering a subshell,
        the server does not process commands from clients for the first one second of the subshell.
        Prior to spec release 6.06.01, the server did not process commands when in a subshell at all.
        """
        self.ready = self._readonly_property("ready", bool)
        """
        status/ready

        .. code-block:: none

            on("change"): Sent when the server is waiting for input at the interactive prompt (True)
                          and after a return has been typed at the interactive prompt (False).
                          The server is available to execute commands from clients when it is ready.
                          
            get: True if the server command thread is busy and unable to immediately process a new command, otherwise False.
        """
        self.simulate = self._readonly_property("simulate", bool)
        """
        status/simulate

        .. code-block:: none
        
            on("change"): Sent when the server enters (True) or leaves (False) simulate mode.
            get: True if server is in simulate mode, otherwise False.

        Note, when in simulate mode, the server will not send commands to hardware devices.
        """

    async def to_str(self) -> str:
        return f"Status(quit={await self.quit.get()}, shell={await self.shell.get()}, ready={await self.ready.get()}, simulate={await self.simulate.get()})"



import pytest
from pyspec.client import Client
from conftest import HOST, PORT
import asyncio


@pytest.mark.asyncio
@pytest.mark.timeout(2)
async def test_client_property_read_write(server_process):
    async with Client(HOST, PORT) as client:
        foo = client._property("foo", int)
        await foo.set(123)
        value = await foo.get()
        assert value == 123


@pytest.mark.asyncio
@pytest.mark.timeout(2)
async def test_client_property_get_next(server_process):
    async with (
        Client(HOST, PORT) as client,
        client._property("ticker", int).subscribed() as ticker,
    ):
        current_value = await ticker.get_next()
        next_value = await ticker.get_next()
        assert current_value + 1 == next_value


@pytest.mark.asyncio
@pytest.mark.timeout(2)
async def test_client_exec_function(server_process):
    async with Client(HOST, PORT) as client:
        result = await client.call("sum", 2, 3)
        assert result == 5


@pytest.mark.asyncio
@pytest.mark.timeout(2)
async def test_client_exec_command(server_process):
    async with Client(HOST, PORT) as client:
        result = await client.exec("2+2")
        assert result == 4


@pytest.mark.asyncio
@pytest.mark.timeout(2)
async def test_client_property_wait_for(server_process):
    async with (
        Client(HOST, PORT) as client,
        client._property("ticker", int).subscribed() as ticker,
        client._property("foo", int).subscribed() as foo,
    ):
        await ticker.set(42)
        async with ticker.wait_for(45):
            pass

        assert await ticker.get() == 45
        await ticker.set(99)
        await ticker.wait_for(105)
        assert await ticker.get() == 105

        async with foo.wait_for(10, timeout=1):
            await foo.set(10)

        with pytest.raises(asyncio.TimeoutError):
            async with foo.wait_for(20, timeout=0.1):
                # We won't set foo to 20, so this should timeout
                pass


@pytest.mark.asyncio
@pytest.mark.timeout(2)
async def test_client_property_get_when_subscribed(server_process):
    async with (
        Client(HOST, PORT) as client,
        client._property("foo", int).subscribed() as foo,
    ):
        await foo.set(123)
        value = await foo.get()
        assert value == 123


@pytest.mark.asyncio
@pytest.mark.timeout(2)
async def test_client_property_get_next_when_subscribed(server_process):
    async with (
        Client(HOST, PORT) as client,
        client._property("ticker", int).subscribed() as ticker,
    ):
        current_value = await ticker.get_next()
        next_value = await ticker.get_next()
        assert next_value == current_value + 1


@pytest.mark.asyncio
@pytest.mark.timeout(2)
async def test_client_boolean_property(server_process):
    async with (
        Client(HOST, PORT) as client,
        client._property("flag", bool).subscribed() as flag,
    ):
        await flag.set(True)
        value = await flag.get()
        assert value is True

        await flag.set(False)
        value = await flag.get()
        assert value is False



from enum import Enum


class Flag(Enum):
    """
    Enumeration of flags that can be sent between the client and server.

    This enum defines all possible flags for the SPEC protocol.
    """

    NONE = 0
    DELETED = 0x1000
    """
    Sent when watched variables or associative array elements are deleted.
    the spec client currently does not take any action on receipt of such events.
    """



from .client import Client, RemoteException
from .server import Server, Property

from . import shared_memory  # type: ignore
from . import file

__all__ = [
    "Client",
    "RemoteException",
    "Property",
    "Server",
    "shared_memory",
    "file",
]



import struct
import re
import sys
from typing import Optional

import numpy as np

re_comment = re.compile(r"(?P<key>\w+)=\"(?P<value>\w+)\"")
re_group = re.compile(r"\[(?P<groupname>.*?)\]")


class ItexImage(object):

    _data: Optional[np.ndarray] = None

    @property
    def data(self) -> np.ndarray:
        if self._data is None:
            raise RuntimeError("No data loaded")
        return self._data

    def __init__(self, filename=None):
        self.head_info = None
        self.comments = {}
        self._data = None

        if filename:
            self.from_file(filename)

    def from_buffer(self, buff):

        _id = struct.unpack("2c", buff[0:2])
        _fst = ""

        for _c in _id:
            _fst += _c.decode("utf-8")

        if _fst != "IM":
            print("not an ITEX data file")
            return

        head_info = struct.unpack("5h", buff[2:12])

        self.width, self.height = head_info[1:3]
        self.xoffset, self.yoffset = head_info[3:5]

        print(self.width, self.height)

        comm_length = head_info[0]
        comm_string = buff[64 : comm_length + 64].decode("utf-8")
        self.parse_comments(comm_string)

        img_size = self.width * self.height * 2
        start_img = comm_length + 64
        end_img = comm_length + 64 + img_size

        data = np.frombuffer(buff[start_img:end_img], dtype=np.uint16)
        self._data = data.reshape(self.height, self.width)
        print(self.data.shape)

        # only if calib present (ScalingXScalingFile in comments)
        try:
            self.calib_data = struct.unpack("672f", buff[end_img:])
        except:
            self.calib_data = None

    def from_file(self, filename):
        buff = open(filename, "rb").read()
        self.from_buffer(buff)

    def get_roisum(self, bx, by, ex, ey):
        pass

    def parse_comments(self, comms):
        curr_group = ""
        cursor = 0
        group_begin = None
        while True:
            mat = re_group.search(comms, cursor)
            if mat:
                if curr_group:
                    group_block = comms[group_begin : mat.start()]
                    print(curr_group)
                    print(group_block)
                curr_group = mat.group("groupname")
                cursor = mat.end()
                group_begin = cursor
            else:
                break

        if group_begin is None:
            raise ValueError("No groups found in comments")

        group_block = comms[group_begin:]
        print(curr_group)
        print(group_block)


def main():
    img = ItexImage(sys.argv[1])
    data = img.data
    print(type(data))
    print(data.shape)
    print(data.mean())


if __name__ == "__main__":
    main()



import ast
import asyncio
import logging
from typing import Any, Callable, Coroutine, TypeVar, Tuple, Union

from pyspec._connection.data import DataType

LOGGER = logging.getLogger("pyspec.server")

SyncOrAsyncCallable = Union[
    Callable[..., DataType], Callable[..., Coroutine[Any, Any, DataType]]
]

F = TypeVar("F", bound=SyncOrAsyncCallable)


def mark_remote_function(function: F) -> F:
    """
    Decorator to mark a function as remotely callable.

    Args:
        function (F): The function to mark as remote.
    Returns:
        F: The same function, marked as remote.
    """
    setattr(function, "_is_remote_function", True)
    setattr(function, "_remote_function_name", function.__name__)
    return function


def is_remote_function(function: SyncOrAsyncCallable) -> bool:
    """
    Check if a function is marked as remotely callable.

    Args:
        function (SyncOrAsyncCallable): The function to check.
    Returns:
        bool: True if the function is marked as remote, False otherwise.
    """
    return getattr(function, "_is_remote_function", False)


def remote_function_name(function: SyncOrAsyncCallable) -> str:
    """
    Get the remote function name for a function.

    Args:
        function (SyncOrAsyncCallable): The function to get the name for.
    Returns:
        str: The remote function name.
    """
    return getattr(function, "_remote_function_name", function.__name__)


def parse_remote_function_string(function_string: str) -> Tuple[str, Tuple[str, ...]]:
    """
    Parse a remote function call string into its name and arguments.

    Args:
        function_string (str): The function call string to parse.
    Returns:
        tuple: Tuple of function name and arguments.
    """
    function_string = function_string.strip()
    if "(" not in function_string or not function_string.endswith(")"):
        raise ValueError(f"Invalid function call string: {function_string}")

    name, args_str = function_string[:-1].split("(", 1)

    args_str = args_str.strip()
    if args_str:
        # Use ast.literal_eval to safely parse the arguments
        # This will handle more complex argument types like strings with commas
        # It is a safer and more constrained alternative to eval
        args = ast.literal_eval(f"({args_str},)")
    else:
        args = ()

    return name, args


def build_remote_function_string(
    function_name: str, args: Tuple[Union[str, float, int], ...]
) -> str:
    """
    Build a remote function call string from its name and arguments.

    Args:
        function_name (str): The function name.
        args (tuple): The arguments to include in the call string.
    Returns:
        str: The remote function call string.
    """
    args_str = ", ".join(repr(arg) for arg in args)
    return f"{function_name}({args_str})"


def remote_function(function: F) -> F:
    # TODO: Need to figure out how to type this properly
    # Since the client will only give you strs.
    """
    Decorator to mark a function as remotely callable.

    Note: All inputs to the function will be received as strings.

    Args:
        function (F): The function to mark as remote.
    Returns:
        F: The same function, marked as remote.
    """
    mark_remote_function(function)
    if not asyncio.iscoroutinefunction(function):
        LOGGER.warning(
            "Remote function '%s' is not asynchronous. "
            "Consider making it async for better performance.",
            function.__name__,
        )
    return function



from pyspec._connection.client_connection import RemoteException

from .client import Client
from ._remote_property_table import (
    Property,
    ReadableProperty,
    WritableProperty,
    EventStream,
)

__all__ = [
    "Client",
    "RemoteException",
    "Property",
    "ReadableProperty",
    "WritableProperty",
    "EventStream",
]
